import dataclasses
import named_arrays as na

__all__ = [
    'AbstractCartesianMatrixArray',
    'AbstractExplicitCartesianMatrixArray',
    'AbstractImplicitCartesianMatrixArray',
    'AbstractCartesianMatrixRandomSample',
    'AbstractCartesianMatrixUniformRandomSample',
    'AbstractCartesianMatrixNormalRandomSample',
    'AbstractParameterizedCartesianMatrixArray',
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianMatrixArray(
    na.AbstractMatrixArray,
    na.AbstractCartesianVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractExplicitCartesianMatrixArray(
    AbstractCartesianMatrixArray,
    na.AbstractExplicitMatrixArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesianMatrixArray(
    AbstractCartesianMatrixArray,
    na.AbstractImplicitMatrixArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianMatrixRandomSample(
    AbstractImplicitCartesianMatrixArray,
    na.AbstractMatrixRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianMatrixUniformRandomSample(
    AbstractCartesianMatrixRandomSample,
    na.AbstractMatrixUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesianMatrixNormalRandomSample(
    AbstractCartesianMatrixRandomSample,
    na.AbstractMatrixNormalRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedCartesianMatrixArray(
    AbstractImplicitCartesianMatrixArray,
    na.AbstractParameterizedMatrixArray,
):
    pass



from __future__ import annotations
from typing import TypeVar, Type, Generic
from typing_extensions import Self
import abc
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractCartesian3dVectorArray',
    'Cartesian3dVectorArray',
    'AbstractImplicitCartesian3dVectorArray',
    'AbstractCartesian3dVectorRandomSample',
    'Cartesian3dVectorUniformRandomSample',
    'Cartesian3dVectorNormalRandomSample',
    'AbstractParameterizedCartesian3dVectorArray',
    'Cartesian3dVectorArrayRange',
    'AbstractCartesian3dVectorSpace',
    'Cartesian3dVectorLinearSpace',
    'Cartesian3dVectorStratifiedRandomSpace',
    'Cartesian3dVectorLogarithmicSpace',
    'Cartesian3dVectorGeometricSpace',
]

XT = TypeVar('XT', bound=na.ArrayLike)
YT = TypeVar('YT', bound=na.ArrayLike)
ZT = TypeVar('ZT', bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorArray(
    na.AbstractCartesian2dVectorArray,
):

    @property
    @abc.abstractmethod
    def z(self: Self) -> na.ArrayLike:
        """
        The `z` component of the vector.
        """

    @property
    def xy(self) -> na.Cartesian2dVectorArray:
        return na.Cartesian2dVectorArray(
            x=self.x,
            y=self.y,
        )

    @property
    def type_abstract(self: Self) -> Type[AbstractCartesian3dVectorArray]:
        return AbstractCartesian3dVectorArray

    @property
    def type_explicit(self: Self) -> Type[Cartesian3dVectorArray]:
        return Cartesian3dVectorArray

    @property
    def type_matrix(self) -> Type[na.Cartesian3dMatrixArray]:
        return na.Cartesian3dMatrixArray

    @property
    def explicit(self) -> Cartesian3dVectorArray:
        return super().explicit

    def cross(
            self,
            other: AbstractCartesian3dVectorArray,
    ) -> na.Cartesian3dVectorArray:
        r"""
        Compute the vector product :math:`\mathbf{a} \times \mathbf{b}`

        Parameters
        ----------
        other
            the right-hand operand of the cross product operation
        """
        a = self
        b = other
        if isinstance(other, na.AbstractCartesian3dVectorArray):
            if len(self.cartesian_nd.components) != 3:
                raise ValueError("all components of `self` must be scalars")
            if len(other.cartesian_nd.components) != 3:
                raise ValueError("all components of `other` must be scalars")
            return self.type_explicit(
                x=+(a.y * b.z - a.z * b.y),
                y=-(a.x * b.z - a.z * b.x),
                z=+(a.x * b.y - a.y * b.x),
            )
        else:
            raise TypeError(
                f"`other` must be an instance of `{na.AbstractCartesian3dVectorArray.__name__}`, "
                f"got `{type(other).__name__}`"
            )


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorArray(
    AbstractCartesian3dVectorArray,
    na.Cartesian2dVectorArray,
    Generic[XT, YT, ZT],
):
    x: XT = 0
    y: YT = 0
    z: ZT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> Cartesian3dVectorArray:
        result = super().from_scalar(scalar, like=like)
        result.z = scalar
        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian3dVectorArray(
    AbstractCartesian3dVectorArray,
    na.AbstractImplicitCartesian2dVectorArray,
):

    @property
    def x(self) -> na.ArrayLike:
        return self.explicit.x

    @property
    def y(self) -> na.ArrayLike:
        return self.explicit.y

    @property
    def z(self) -> na.ArrayLike:
        return self.explicit.z


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorRandomSample(
    AbstractImplicitCartesian3dVectorArray,
    na.AbstractCartesian2dVectorRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorUniformRandomSample(
    AbstractCartesian3dVectorRandomSample,
    na.Cartesian2dVectorUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorNormalRandomSample(
    AbstractCartesian3dVectorRandomSample,
    na.Cartesian2dVectorNormalRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedCartesian3dVectorArray(
    AbstractImplicitCartesian3dVectorArray,
    na.AbstractParameterizedCartesian2dVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorArrayRange(
    AbstractParameterizedCartesian3dVectorArray,
    na.Cartesian2dVectorArrayRange,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorSpace(
    AbstractParameterizedCartesian3dVectorArray,
    na.AbstractCartesian2dVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorLinearSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorStratifiedRandomSpace(
    Cartesian3dVectorLinearSpace,
    na.Cartesian2dVectorStratifiedRandomSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorLogarithmicSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorLogarithmicSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorGeometricSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorGeometricSpace,
):
    pass



from __future__ import annotations
from typing import Literal
import matplotlib.axes
import matplotlib.figure
import matplotlib.artist
import matplotlib.pyplot as plt
import numpy.typing as npt
import named_arrays as na

__all__ = [
    "subplots",
    "plot",
    "fill",
]


def subplots(
        axis_rows: str = "subplots_row",
        ncols: int = 1,
        axis_cols: str = "subplots_col",
        nrows: int = 1,
        *,
        sharex: bool | Literal["none", "all", "row", "col"] = False,
        sharey: bool | Literal["none", "all", "row", "col"] = False,
        squeeze: bool = True,
        **kwargs,
) -> tuple[
    matplotlib.figure.Figure,
    matplotlib.axes.Axes | na.ScalarArray[npt.NDArray],
]:
    """
    A thin wrapper around :func:`matplotlib.pyplot.subplots()` which allows for
    providing axis names to the rows and columns.

    Parameters
    ----------
    axis_rows
        Name of the axis representing the rows in the subplot grid.
        If :obj:`None`, the ``squeeze`` argument must be :obj:`True`.
    nrows
        Number of rows in the subplot grid
    axis_cols
        Name of the axis representing the columns in the subplot grid
        If :obj:`None`, the ``squeeze`` argument must be :obj:`True`.
    ncols
        Number of columns in the subplot grid
    sharex
        Controls whether all the :class:`matplotlib.axes.Axes` instances share the same horizontal axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    sharey
        Controls whether all the :class`matplotlib.axes.Axes` instances share the same vertical axis properties.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    squeeze
        If :obj:`True`, :func:`numpy.squeeze` is called on the result, which removes singleton dimensions from the
        array.
        See the documentation of :func:`matplotlib.pyplot.subplots` for more information.
    kwargs
        Additional keyword arguments passed to :func:`matplotlib.pyplot.subplots`
    """

    shape = {axis_rows: nrows, axis_cols: ncols}

    if squeeze:
        shape = {axis: shape[axis] for axis in shape if shape[axis] != 1}

    fig, axs = plt.subplots(
        ncols=ncols,
        nrows=nrows,
        sharex=sharex,
        sharey=sharey,
        squeeze=squeeze,
        **kwargs,
    )

    return fig, na.ScalarArray(axs, axes=tuple(shape.keys()))


def plot(
        *args: na.AbstractScalar,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalar = True,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:
    """
    A thin wrapper around :meth:`matplotlib.axes.Axes.plot` for named arrays.

    The main difference of this function from :func:`matplotlib.pyplot.plot` is the addition of the ``axis`` parameter
    indicating along which axis the lines should be connected.

    Parameters
    ----------
    args
        Same signature as :meth:`matplotlib.axes.Axes.plot`.
        If ``ax`` is a 2D plot, ``*args`` should be ``y`` or ``x, y``.
        If ``ax`` is a 3D plot, ``*args`` should be ``x, y`` or ``x, y, z``.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    axis
        The name of the axis that the plot lines should be connected along.
        If :obj:`None`, the broadcasted shape of ``args`` should have only one element,
        otherwise a :class:`ValueError` is raised.
    where
        A boolean array that selects which elements to plot
    transformation
        A callable that is applied to args before plotting
    components
        The component names of ``*args`` to plot, helpful if ``*args`` are an instance of
        :class:`named_arrays.AbstractVectorArray`.
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.plot`.
        These can be instances of :class:`named_arrays.AbstractArray`.

    Returns
    -------
        An array of artists that were plotted

    Examples
    --------

    Plot a single scalar

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        x = na.linspace(0, 2 * np.pi, axis="x",  num=101)
        y = np.sin(x)

        plt.figure();
        na.plt.plot(x, y);

    Plot an array of scalars

    .. jupyter-execute::

        z = na.linspace(0, np.pi, axis="z", num=5)

        y = np.sin(x - z)

        plt.figure();
        na.plt.plot(x, y, axis="x");

    Plot an uncertain scalar

    .. jupyter-execute::

        ux = na.NormalUncertainScalarArray(x, width=0.2)
        uy = np.sin(ux)

        plt.figure();
        na.plt.plot(x, uy);

    Broadcast an array of scalars against an array of :class:`matplotlib.axes.Axes`

    .. jupyter-execute::

        fig, ax = na.plt.subplots(axis_rows="z", nrows=z.shape["z"], sharex=True)

        na.plt.plot(x, y, ax=ax, axis="x");

    Plot a 2D Cartesian vector

    .. jupyter-execute::

        v = na.Cartesian2dVectorArray(x, np.sin(x))

        plt.figure()
        na.plt.plot(v)

    """
    if transformation is not None:
        args = tuple(transformation(arg) for arg in args)
    return na._named_array_function(
        plot,
        *args,
        ax=ax,
        axis=axis,
        where=where,
        components=components,
        **kwargs,
    )


def fill(
        *args: na.AbstractArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalar = True,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray]:
    """
    Plot filled polygons

    This is a thin wrapper around :meth:`matplotlib.axes.Axes.fill` for named arrays.

    The main difference of this function from :func:`matplotlib.pyplot.fill` is the addition of the ``axis`` parameter
    indicating along which axis the lines should be connected.

    Parameters
    ----------
    args
        Same signature as :meth:`matplotlib.axes.Axes.fill`.
        If ``ax`` is a 2D plot, ``*args`` should be ``x, y``.
        If ``ax`` is a 3D plot, ``*args`` should be ``x, y, z``.
    ax
        The instances of :class:`matplotlib.axes.Axes` to use.
        If :obj:`None`, calls :func:`matplotlib.pyplot.gca` to get the current axes.
        If an instance of :class:`named_arrays.ScalarArray`, ``ax.shape`` should be a subset of the broadcasted shape of
        ``*args``.
    axis
        The name of the axis that the plot lines should be connected along.
        If :obj:`None`, the broadcasted shape of ``args`` should have only one element,
        otherwise a :class:`ValueError` is raised.
    where
        A boolean array that selects which elements to plot
    transformation
        A callable that is applied to args before plotting
    components
        The component names of ``*args`` to plot, helpful if ``*args`` are an instance of
        :class:`named_arrays.AbstractVectorArray`.
    kwargs
        Additional keyword arguments passed to :meth:`matplotlib.axes.Axes.fill`.
        These can be instances of :class:`named_arrays.AbstractArray`.
    """
    if transformation is not None:
        args = tuple(transformation(arg) for arg in args)
    return na._named_array_function(
        fill,
        *args,
        ax=ax,
        axis=axis,
        where=where,
        components=components,
        **kwargs,
    )


from typing import Callable
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.scalar_named_array_functions

__all__ = [
    "ASARRAY_LIKE_FUNCTIONS",
]

ASARRAY_LIKE_FUNCTIONS = named_arrays._scalars.scalar_named_array_functions.ASARRAY_LIKE_FUNCTIONS
HANDLED_FUNCTIONS = dict()

def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def asarray_like(
        func: Callable,
        a: None | float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray | na.AbstractFunctionArray,
        dtype: None | type | np.dtype = None,
        order: None | str = None,
        *,
        like: None | float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray | na.AbstractFunctionArray = None,
) -> None | na.AbstractFunctionArray:

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractFunctionArray):
            a_inputs = a.inputs
            a_outputs = a.outputs
        elif isinstance(a, na.AbstractVectorArray):
            a_inputs = a_outputs = a
        elif isinstance(a, na.AbstractScalar):
            a_inputs = a_outputs = a
        else:
            return NotImplemented
    else:
        a_inputs = a_outputs = a

    if isinstance(like, na.AbstractArray):
        if isinstance(like, na.AbstractFunctionArray):
            like_inputs = like.inputs
            like_outputs = like.outputs
            type_like = like.type_explicit
        elif isinstance(like, na.AbstractVectorArray):
            like_inputs = like_outputs = like
            type_like = na.FunctionArray
        elif isinstance(like, na.AbstractScalar):
            like_inputs = like_outputs = like
            type_like = na.FunctionArray
        else:
            return NotImplemented
    else:
        like_inputs = like_outputs = like
        type_like = na.FunctionArray

    return type_like(
        inputs=func(
            a=a_inputs,
            dtype=dtype,
            order=order,
            like=like_inputs,
        ),
        outputs=func(
            a=a_outputs,
            dtype=dtype,
            order=order,
            like=like_outputs,
        ),
    )


@_implements(na.unit)
def unit(
        a: na.AbstractFunctionArray,
        squeeze: bool = True,
) -> None | u.UnitBase | na.AbstractArray:
    return na.unit(a.outputs, squeeze=squeeze)


@_implements(na.unit_normalized)
def unit_normalized(
        a: na.AbstractFunctionArray,
        squeeze: bool = True,
) -> u.UnitBase | na.AbstractArray:
    return na.unit_normalized(a.outputs, squeeze=squeeze)



from __future__ import annotations
from typing import Sequence, Callable, Type
import numpy as np
import astropy.units as u
import named_arrays as na
from . import scalars

__all__ = [
    'SINGLE_ARG_FUNCTIONS',
    'ARRAY_CREATION_LIKE_FUNCTIONS',
    'DEFAULT_FUNCTIONS',
    'PERCENTILE_LIKE_FUNCTIONS',
    'ARG_REDUCE_FUNCTIONS',
    'FFT_LIKE_FUNCTIONS',
    'FFTN_LIKE_FUNCTIONS',
    'HANDLED_FUNCTIONS',
]
SINGLE_ARG_FUNCTIONS = [
    np.real,
    np.imag,
]
ARRAY_CREATION_LIKE_FUNCTIONS = [
    np.empty_like,
    np.zeros_like,
    np.ones_like,
]
SEQUENCE_FUNCTIONS = [
    np.linspace,
    np.logspace,
    np.geomspace,
]
DEFAULT_FUNCTIONS = [
    np.ndim,
    np.min,
    np.nanmin,
    np.max,
    np.nanmax,
    np.sum,
    np.nansum,
    np.prod,
    np.nanprod,
    np.mean,
    np.nanmean,
    np.std,
    np.nanstd,
    np.var,
    np.nanvar,
    np.median,
    np.nanmedian,
    np.all,
    np.any,
    np.ptp,
    np.count_nonzero,
]
PERCENTILE_LIKE_FUNCTIONS = [
    np.percentile,
    np.nanpercentile,
    np.quantile,
    np.nanquantile,
]
ARG_REDUCE_FUNCTIONS = [
    np.argmin,
    np.nanargmin,
    np.argmax,
    np.nanargmax,
]
FFT_LIKE_FUNCTIONS = [
    np.fft.fft,
    np.fft.ifft,
    np.fft.rfft,
    np.fft.irfft,
]
FFTN_LIKE_FUNCTIONS = [
    np.fft.fft2,
    np.fft.ifft2,
    np.fft.rfft2,
    np.fft.irfft2,
    np.fft.fftn,
    np.fft.ifftn,
    np.fft.rfftn,
    np.fft.irfftn,
]
HANDLED_FUNCTIONS = dict()


def array_function_single_arg(
    func: Callable,
    a: na.AbstractScalarArray,
) -> na.ScalarArray:
    return a.type_explicit(
        ndarray=func(a.ndarray),
        axes=a.axes,
    )


def array_function_array_creation_like(
        func: Callable,
        prototype: na.AbstractScalarArray,
        dtype: None | type | np.dtype = None,
        order: str = "K",
        subok: bool = True,
        shape: dict[str, int] = None,
):
    if shape is None:
        shape = prototype.shape

    return prototype.type_explicit(
        ndarray=func(
            prototype.ndarray,
            dtype=dtype,
            order=order,
            subok=subok,
            shape=tuple(shape.values()),
        ),
        axes=tuple(shape.keys()),
    )


def array_function_sequence(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalarArray,
        axis: str,
        num: int = 50,
        endpoint: bool = True,
        dtype: None | type | np.dtype = None,
        **kwargs: float | u.Quantity | na.AbstractScalarArray,
):
    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    shape = na.shape_broadcasted(*args, *kwargs.values())

    args = tuple(arg.ndarray_aligned(shape) for arg in args)
    kwargs = {k: kwargs[k].ndarray_aligned(shape) for k in kwargs}

    if isinstance(axis, na.AbstractArray):
        if isinstance(axis, na.AbstractScalarArray):
            axis = axis.ndarray
        else:
            return NotImplemented

    return na.ScalarArray(
        ndarray=func(
            *args,
            num=num,
            endpoint=endpoint,
            dtype=dtype,
            axis=~0,
            **kwargs,
        ),
        axes=tuple(shape) + (axis, ),
    )


def array_function_default(
        func: Callable,
        a: na.AbstractScalarArray,
        axis: None | str | Sequence[str] = None,
        dtype: type | np.dtype = np._NoValue,
        out: None | na.ScalarArray = None,
        keepdims: bool = False,
        initial: bool | int | float | complex | u.Quantity = np._NoValue,
        where: bool | na.AbstractScalarArray = np._NoValue,
):
    a = a.explicit
    shape = na.shape_broadcasted(a, where)

    axis_normalized = na.axis_normalized(a, axis=axis)

    if axis is not None:
        if not set(axis_normalized).issubset(shape):
            raise ValueError(
                f"the `axis` argument must be `None` or a subset of the broadcasted shape of `a` and `where`, "
                f"got {axis} for `axis`, but `{shape} for `shape`"
            )

    if out is not None:
        if not isinstance(out, na.ScalarArray):
            raise ValueError(f"`out` should be `None` or an instance of `{a.type_explicit}`, got `{type(out)}`")
        axes_ndarray = out.axes
        if not keepdims:
            axes_ndarray = axes_ndarray + tuple(ax for ax in shape if ax not in out.axes)
    else:
        axes_ndarray = tuple(shape)

    kwargs = dict()
    kwargs["axis"] = tuple(axes_ndarray.index(ax) for ax in axis_normalized)
    if dtype is not np._NoValue:
        kwargs["dtype"] = dtype
    if out is not None and isinstance(out.ndarray, np.ndarray):
        kwargs["out"] = out.ndarray
    else:
        kwargs["out"] = None
    kwargs["keepdims"] = keepdims
    if initial is not np._NoValue:
        kwargs["initial"] = initial
    if where is not np._NoValue:
        if isinstance(where, na.AbstractArray):
            if isinstance(where, na.AbstractScalarArray):
                kwargs["where"] = where.ndarray_aligned(axes_ndarray)
            else:
                return NotImplemented
        else:
            kwargs["where"] = where

    axes_result = tuple(ax for ax in axes_ndarray if ax not in axis_normalized) if not keepdims else axes_ndarray

    result_ndarray = func(a.ndarray_aligned(axes_ndarray), **kwargs)

    if out is None:
        result = na.ScalarArray(
            ndarray=result_ndarray,
            axes=axes_result,
        )
    else:
        out.ndarray = result_ndarray
        result = out
    return result


def array_function_percentile_like(
        func: Callable,
        a: na.AbstractScalarArray,
        q: float | u.Quantity | na.AbstractScalarArray,
        axis: None | str | Sequence[str] = None,
        out: None | na.ScalarArray = None,
        overwrite_input: bool = np._NoValue,
        method: str = np._NoValue,
        keepdims: bool = False,
) -> na.ScalarArray:

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalarArray):
            a = a.explicit
        else:
            return NotImplemented
    else:
        a = na.ScalarArray(a)

    axes_a = a.axes

    axis_normalized = na.axis_normalized(a, axis=axis)

    if axis is not None:
        if not set(axis_normalized).issubset(axes_a):
            raise ValueError(
                f"the `axis` argument must be `None` or a subset of `a.axes`, "
                f"got {axis} for `axis`, but `{a.axes} for `a.axes`"
            )

    if isinstance(q, na.AbstractArray):
        if isinstance(q, na.AbstractScalarArray):
            q = q.explicit
        else:
            return NotImplemented
    else:
        q = na.ScalarArray(q)

    axes_q = q.axes

    axis_union = set(a.axes) & set(q.axes)
    if axis_union:
        raise ValueError(f"'q' must not have any shared axes with 'a', but axes {axis_union} are shared")

    axes_result = axes_q + (tuple(ax for ax in axes_a if ax not in axis_normalized) if not keepdims else axes_a)

    kwargs = dict()
    kwargs["axis"] = tuple(axes_a.index(ax) for ax in axis_normalized)

    if out is not None:
        if not isinstance(out, na.ScalarArray):
            raise ValueError(f"`out` should be `None` or an instance of `named_arrays.ScalarArray`, got `{type(out)}`")
        kwargs["out"] = out.ndarray_aligned(axes_result)
    else:
        kwargs["out"] = out

    if overwrite_input is not np._NoValue:
        kwargs['overwrite_input'] = overwrite_input
    if method is not np._NoValue:
        kwargs['method'] = method
    kwargs['keepdims'] = keepdims

    result_ndarray = func(a.ndarray, q.ndarray, **kwargs)

    result = na.ScalarArray(
        ndarray=result_ndarray,
        axes=axes_result,
    )

    if out is not None:
        out.ndarray = result.ndarray_aligned(out.shape)
        result = out
    return result


def array_function_arg_reduce(
        func: Callable,
        a: na.AbstractScalarArray,
        axis: None | str | Sequence[str] = None,
) -> dict[str, na.ScalarArray]:

    a = a.explicit
    shape_a = a.shape

    if axis is not None:
        if isinstance(axis, str):
            axis = (axis, )
        if not set(axis).issubset(a.axes):
            raise ValueError(f"Reduction axes {axis} are not a subset of the array axes {a.axes}")
    else:
        if not a.shape:
            raise ValueError(f"Applying {func} to zero-dimensional arrays is not supported")

    if axis is not None:
        axis_flattened = na.flatten_axes(axis)
        a_flattened = a.combine_axes(axes=axis, axis_new=axis_flattened)
        axis_ndarray = a_flattened.axes.index(axis_flattened)

        indices = na.ScalarArray(
            ndarray=func(a_flattened.ndarray, axis=axis_ndarray),
            axes=tuple(ax for ax in a_flattened.axes if ax != axis_flattened),
        )

        result = np.unravel_index(indices, shape={ax: shape_a[ax] for ax in axis})
        for ax in shape_a:
            if ax not in axis:
                result[ax] = na.ScalarArrayRange(0, shape_a[ax], axis=ax)

    else:
        index = na.ScalarArray(func(a.ndarray, axis=axis))
        result = np.unravel_index(index, shape=shape_a)

    return result


def array_function_fft_like(
        func: Callable,
        a: na.AbstractScalarArray,
        axis: tuple[str, str],
        n: None | int = None,
        norm: str = "backward"
) -> na.ScalarArray:

    return na.ScalarArray(
        ndarray=func(
            a=a.ndarray,
            n=n,
            axis=a.axes.index(axis[0]),
            norm=norm,
        ),
        axes=tuple(axis[1] if ax == axis[0] else ax for ax in a.axes),
    )


def array_function_fftn_like(
        func: Callable,
        a: na.AbstractScalarArray,
        axes: dict[str, str],
        s: None | dict[str, int] = None,
        norm: str = "backward",
) -> na.ScalarArray:

    a = a.explicit
    shape = a.shape

    if not all(ax in shape for ax in axes):
        raise ValueError(f"Not all transform axes {axes} are in input array shape {shape}")

    if s is None:
        s = {ax: shape[ax] for ax in axes}

    if not axes.keys() == s.keys():
        raise ValueError(f"'axes' {axes} and 's' {s} must have the same keys")

    return na.ScalarArray(
        ndarray=func(
            a=a.ndarray,
            s={ax: s[ax] for ax in axes}.values(),
            axes=tuple(a.axes.index(ax) for ax in axes),
            norm=norm,
        ),
        axes=tuple(axes[ax] if ax in axes else ax for ax in a.axes),
    )


def implements(numpy_function: Callable):
    """Register an __array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func
    return decorator


@implements(np.copyto)
def copyto(
        dst: na.ScalarArray,
        src: na.AbstractScalarArray,
        casting: str = "same_kind",
        where: bool | na.AbstractScalarArray = True,
):

    if not isinstance(dst, na.ScalarArray):
        return NotImplemented

    shape = dst.shape

    if isinstance(src, na.AbstractArray):
        if isinstance(src, na.AbstractScalarArray):
            src_ndarray = src.ndarray_aligned(shape)
        else:
            return NotImplemented
    else:
        src_ndarray = src

    if isinstance(where, na.AbstractArray):
        if isinstance(where, na.AbstractScalarArray):
            where_ndarray = where.ndarray_aligned(shape)
        else:
            return NotImplemented
    else:
        where_ndarray = where

    try:
        np.copyto(
            dst=dst.ndarray,
            src=src_ndarray,
            casting=casting,
            where=where_ndarray,
        )
    except TypeError:
        dst.ndarray = src.ndarray


@implements(np.broadcast_to)
def broadcast_to(
        array: na.AbstractScalarArray,
        shape: dict[str, int],
):
    return na.ScalarArray(
        ndarray=np.broadcast_to(array.ndarray_aligned(shape), tuple(shape.values()), subok=True),
        axes=tuple(shape.keys()),
    )


@implements(np.transpose)
def transpose(
        a: na.AbstractScalarArray,
        axes: None | Sequence[str] = None,
) -> na.ScalarArray:

    if axes is not None:
        a = a.add_axes(axes)
    axes = tuple(reversed(a.axes) if axes is None else axes)
    return na.ScalarArray(
        ndarray=np.transpose(
            a=a.ndarray,
            axes=tuple(a.axes.index(ax) for ax in axes),
        ),
        axes=axes,
    )


@implements(np.moveaxis)
def moveaxis(
        a: na.AbstractScalarArray,
        source: str | Sequence[str],
        destination: str | Sequence[str],
) -> na.ScalarArray:

    axes = list(a.axes)

    if isinstance(source, str):
        source = (source,)
    if isinstance(destination, str):
        destination = (destination,)

    set_axis_diff = set(source) - set(axes)
    if set_axis_diff:
        raise ValueError(f"source axes {tuple(set_axis_diff)} not in array axes {axes}")

    for src, dest in zip(source, destination):
        axes[axes.index(src)] = dest

    return na.ScalarArray(
        ndarray=a.ndarray,
        axes=tuple(axes)
    )


@implements(np.reshape)
def reshape(a: na.AbstractScalarArray, newshape: dict[str, int]) -> na.ScalarArray:
    return na.ScalarArray(
        ndarray=np.reshape(a.ndarray, tuple(newshape.values())),
        axes=tuple(newshape.keys()),
    )


@implements(np.stack)
def stack(
        arrays: Sequence[bool | int | float | complex | str | u.Quantity | na.AbstractScalarArray],
        axis: str,
        out: None | na.ScalarArray = None,
        *,
        dtype: str | np.dtype | Type = None,
        casting: str = "same_kind",
) -> na.ScalarArray:
    arrays = [na.ScalarArray(arr) if not isinstance(arr, na.AbstractArray) else arr for arr in arrays]
    for array in arrays:
        if not isinstance(array, na.AbstractScalarArray):
            return NotImplemented
    shape = na.shape_broadcasted(*arrays)
    if axis in shape:
        raise ValueError(f"axis '{axis}' already in array")
    arrays = [arr.broadcast_to(shape).ndarray for arr in arrays]

    axes_new = (axis,) + tuple(shape.keys())
    axis_ndarray = 0

    if out is not None:
        np.stack(
            arrays=arrays,
            axis=axis_ndarray,
            out=out.transpose(axes_new).ndarray,
            dtype=dtype,
            casting=casting
        )
        return out
    else:
        return na.ScalarArray(
            ndarray=np.stack(
                arrays=arrays,
                axis=axis_ndarray,
                out=out,
                dtype=dtype,
                casting=casting,
            ),
            axes=axes_new,
        )


@implements(np.concatenate)
def concatenate(
        arrays: Sequence[bool | int | float | complex | str | u.Quantity | na.AbstractScalarArray],
        axis: str,
        out: None | na.ScalarArray = None,
        dtype: None | str | Type | np.dtype = None,
        casting: str = "same_kind",
) -> na.ScalarArray:

    arrays = [na.ScalarArray(arr) if not isinstance(arr, na.AbstractArray) else arr.explicit for arr in arrays]
    for array in arrays:
        if not isinstance(array, na.AbstractScalarArray):
            return NotImplemented

    if any(axis not in array.shape for array in arrays):
        raise ValueError(f"axis '{axis}' must be present in all the input arrays, got {[a.axes for a in arrays]}")

    shapes = []
    for array in arrays:
        shape = array.shape
        shape[axis] = 1
        shapes.append(shape)

    shape_prototype = na.broadcast_shapes(*shapes)

    ndarrays = []
    for array in arrays:
        shape = shape_prototype.copy()
        if axis in array.axes:
            shape[axis] = array.shape[axis]
        array = array.broadcast_to(shape)
        ndarrays.append(array.ndarray)

    axis_index = tuple(shape_prototype).index(axis)

    if out is not None:
        np.concatenate(
            ndarrays,
            axis=axis_index,
            out=out.transpose(tuple(shape_prototype.keys())).ndarray,
            dtype=dtype,
            casting=casting,
        )
        return out
    else:
        return na.ScalarArray(
            ndarray=np.concatenate(
                ndarrays,
                axis=axis_index,
                out=out,
                dtype=dtype,
                casting=casting
            ),
            axes=tuple(shape_prototype.keys()),
        )


@implements(np.sort)
def sort(
        a: na.AbstractScalarArray,
        axis: None | str | Sequence[str],
        kind: None | str = None,
        order: None | str | list[str] = None,
) -> na.ScalarArray:

    a = a.explicit

    if axis is not None:
        if isinstance(axis, str):
            axis = (axis, )
        else:
            axis = tuple(axis)

        if not axis:
            raise ValueError(f"if `axis` is a sequence, it must not be empty, got {axis}")

        if not set(axis).issubset(a.axes):
            raise ValueError(f"`axis`, {axis} is not a subset of `a.axes`, {a.axes}")

        axis_new = na.flatten_axes(axis)
        a = a.combine_axes(axes=axis, axis_new=axis_new)
        axis = axis_new

    else:
        if not a.shape:
            return a

    return na.ScalarArray(
        ndarray=np.sort(
            a=a.ndarray,
            axis=a.axes.index(axis) if axis is not None else axis,
            kind=kind,
            order=order
        ),
        axes=a.axes if axis is not None else (a.axes_flattened, )
    )


@implements(np.argsort)
def argsort(
        a: na.AbstractScalarArray,
        axis: None | str,
        kind: None | str = None,
        order: None | str | list[str] = None,
) -> dict[str, na.ScalarArray]:

    a = a.explicit
    shape_a = a.shape

    if axis is not None:
        if isinstance(axis, str):
            axis = (axis, )
        else:
            axis = tuple(axis)

        if not axis:
            raise ValueError(f"if `axis` is a sequence, it must not be empty, got {axis}")

        if not set(axis).issubset(a.axes):
            raise ValueError(f"`axis`, {axis} is not a subset of `a.axes`, {a.axes}")

    else:
        if not a.shape:
            return dict()
        else:
            axis = a.axes

    axis_flattened = na.flatten_axes(axis)
    a_flattened = a.combine_axes(axes=axis, axis_new=axis_flattened)

    indices_ndarray = np.argsort(
        a=a_flattened.ndarray,
        axis=a_flattened.axes.index(axis_flattened),
        kind=kind,
        order=order,
    )
    indices = na.ScalarArray(indices_ndarray, axes=a_flattened.axes)

    result = np.unravel_index(indices, shape={ax: shape_a[ax] for ax in axis})

    for ax in shape_a:
        if ax not in result:
            result[ax] = na.ScalarArrayRange(
                start=0,
                stop=shape_a[ax],
                axis=ax,
            )

    return result


@implements(np.unravel_index)
def unravel_index(indices: na.AbstractScalarArray, shape: dict[str, int]) -> dict[str, na.ScalarArray]:

    result_ndarray = np.unravel_index(indices=indices.ndarray, shape=tuple(shape.values()))
    result = dict()  # type: dict[str, na.ScalarArray]
    for axis, ndarray in zip(shape, result_ndarray):
        result[axis] = na.ScalarArray(
            ndarray=ndarray,
            axes=indices.axes,
        )
    return result


@implements(np.array_equal)
def array_equal(
        a1: na.AbstractScalarArray,
        a2: na.AbstractScalarArray,
        equal_nan: bool = False,
) -> bool:
    return np.array_equal(
        a1=na.as_named_array(a1).ndarray,
        a2=na.as_named_array(a2).ndarray,
        equal_nan=equal_nan,
    )


@implements(np.array_equiv)
def array_equiv(
        a1: na.AbstractScalarArray,
        a2: na.AbstractScalarArray,
) -> bool:
    shape = na.shape_broadcasted(a1, a2)
    return np.array_equiv(
        a1=na.as_named_array(a1).ndarray_aligned(shape),
        a2=na.as_named_array(a2).ndarray_aligned(shape),
    )


@implements(np.allclose)
def allclose(
        a: na.ScalarLike,
        b: na.ScalarLike,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
) -> bool:

    shape = na.shape_broadcasted(a, b)

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalarArray):
            a = a.ndarray_aligned(shape)
        else:
            return NotImplemented

    if isinstance(b, na.AbstractArray):
        if isinstance(b, na.AbstractScalarArray):
            b = b.ndarray_aligned(shape)
        else:
            return NotImplemented

    return np.allclose(
        a=a,
        b=b,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
    )


@implements(np.nonzero)
def nonzero(a: na.AbstractScalarArray):
    a = a.explicit
    if not a.shape:
        return dict()
    axes = a.axes
    axes_flattened = a.axes_flattened
    result = np.nonzero(a.ndarray)
    return {axes[r]: na.ScalarArray(result[r], axes=(axes_flattened, )) for r, _ in enumerate(result)}


@implements(np.where)
def where(
    condition: na.AbstractScalarArray,
    x: float | u.Quantity | na.AbstractScalarArray,
    y: float | u.Quantity | na.AbstractScalarArray,
) -> na.ScalarArray:
    try:
        condition = scalars._normalize(condition)
        x = scalars._normalize(x)
        y = scalars._normalize(y)
    except scalars.ScalarTypeError:
        return NotImplemented

    shape = na.shape_broadcasted(condition, x, y)

    return condition.type_explicit(
        ndarray=np.where(
            condition.ndarray_aligned(shape),
            x.ndarray_aligned(shape),
            y.ndarray_aligned(shape),
        ),
        axes=tuple(shape),
    )


@implements(np.nan_to_num)
def nan_to_num(
        x: na.AbstractScalarArray,
        copy: bool = True,
        nan: float = 0.0,
        posinf: None | float = None,
        neginf: None | float = None,
):
    result_ndarray = np.nan_to_num(x.ndarray, copy=copy, nan=nan, posinf=posinf, neginf=neginf)
    if copy:
        return na.ScalarArray(
            ndarray=result_ndarray,
            axes=x.axes,
        )
    else:
        if not isinstance(x, na.AbstractExplicitArray):
            raise ValueError("can't write to an array that is not an instance of `named_array.AbstractExplictArray`")
        return x


@implements(np.convolve)
def convolve(
        a: na.ScalarLike,
        v: na.ScalarLike,
        mode: str = 'full',
) -> na.AbstractScalarArray:

    if not isinstance(a, na.AbstractArray):
        a = na.ScalarArray(a)
    if not isinstance(v, na.AbstractArray):
        v = na.ScalarArray(v)

    if a.ndarray is None or v.ndarray is None:
        return na.ScalarArray(None)

    shape_broadcasted = na.shape_broadcasted(a, v)
    if len(shape_broadcasted) > 1:
        raise ValueError(f"'a' and 'v' must broadcast to a 1D shape, got {shape_broadcasted}")

    result_ndarray = np.convolve(a.ndarray, v.ndarray, mode=mode)
    if len(shape_broadcasted) == 0:
        result_ndarray = result_ndarray[0]

    return na.ScalarArray(
        ndarray=result_ndarray,
        axes=tuple(shape_broadcasted.keys()),
    )



from typing import Type, Callable, Sequence
import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays.tests.test_core
import named_arrays._vectors.tests.test_vectors
from . import test_vectors_cartesian

__all__ = [
    'AbstractTestAbstractCartesian2dVectorArray',
    'TestCartesian2dVectorArray',
    'TestCartesian2dVectorArrayCreation',
    'AbstractTestAbstractImplicitCartesian2dVectorArray',
    'AbstractTestAbstractCartesian2dVectorRandomSample',
    'TestCartesian2dVectorUniformRandomSample',
    'TestCartesian2dVectorNormalRandomSample',
    'AbstractTestAbstractParameterizedCartesian2dVectorArray',
    'TestCartesian2dVectorArrayRange',
    'AbstractTestAbstractCartesian2dVectorSpace',
    'TestCartesian2dVectorLinearSpace',
]

_num_x = test_vectors_cartesian._num_x
_num_y = test_vectors_cartesian._num_y
_num_z = test_vectors_cartesian._num_z
_num_distribution = test_vectors_cartesian._num_distribution


def _cartesian2d_arrays():
    arrays_numeric_x = [
        4,
        na.ScalarUniformRandomSample(-4, 4, shape_random=dict(y=_num_y)),
        na.Cartesian2dVectorArray(x=2, y=3)
    ]
    units_x = [1, u.mm]
    arrays_numeric_x = [a * unit for a in arrays_numeric_x for unit in units_x]

    arrays_numeric_y = [
        5.,
        na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
        na.UniformUncertainScalarArray(
            nominal=na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
            width=1,
            num_distribution=_num_distribution
        ),
        na.Cartesian2dVectorArray(x=4, y=8)
    ]
    units_y = [1, u.mm]
    arrays_numeric_y = [a * unit for a in arrays_numeric_y for unit in units_y]

    arrays = [na.Cartesian2dVectorArray(x=ax, y=ay) for ax in arrays_numeric_x for ay in arrays_numeric_y]
    return arrays


def _cartesian2d_arrays_2():
    units = [1, u.mm]
    arrays_scalar = [
        6,
        na.ScalarArray(6),
        na.ScalarUniformRandomSample(-6, 6, shape_random=dict(y=_num_y, x=_num_x)),
        na.UniformUncertainScalarArray(
            nominal=na.ScalarUniformRandomSample(-6, 6, shape_random=dict(y=_num_y, x=_num_x)),
            width=1,
            num_distribution=_num_distribution
        )
    ]
    arrays_scalar = [a * unit for a in arrays_scalar for unit in units]
    arrays_vector_x = [
        7,
        na.ScalarUniformRandomSample(-7, 7, shape_random=dict(y=_num_y)),
    ]
    arrays_vector_x = [a * unit for a in arrays_vector_x for unit in units]
    arrays_vector_y = [
        8,
        na.ScalarUniformRandomSample(-8, 8, shape_random=dict(y=_num_y, x=_num_x)),
    ]
    arrays_vector_y = [a * unit for a in arrays_vector_y for unit in units]
    arrays_vector = [na.Cartesian2dVectorArray(x=ax, y=ay) for ax in arrays_vector_x for ay in arrays_vector_y]
    arrays = arrays_scalar + arrays_vector
    return arrays


def _cartesian2d_items() -> list[na.AbstractArray | dict[str, int, slice, na.AbstractArray]]:
    return [
            dict(y=0),
            dict(y=slice(0, 1)),
            dict(y=na.ScalarArrayRange(0, 2, axis='y')),
            dict(
                y=na.Cartesian2dVectorArray(
                    x=na.ScalarArrayRange(0, 2, axis='y'),
                    y=na.ScalarArrayRange(0, 2, axis='y'),
                )
            ),
            dict(
                y=na.UncertainScalarArray(
                    nominal=na.ScalarArray(np.array([0, 1]), axes=('y',)),
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0, ], [1, ]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                ),
                _distribution=na.UncertainScalarArray(
                    nominal=None,
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0], [0]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                )
            ),
            na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
            na.UniformUncertainScalarArray(
                nominal=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                width=0.1,
                num_distribution=_num_distribution,
            ) > 0.5,
            na.Cartesian2dVectorArray(
                x=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.3,
                y=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
            ),
        ]

class AbstractTestAbstractCartesian2dVectorArray(
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray,
):

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=_cartesian2d_items()
    )
    def test__getitem__(
            self,
            array: na.AbstractCartesian2dVectorArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize('array_2', _cartesian2d_arrays_2())
    class TestUfuncBinary(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize('array_2', _cartesian2d_arrays_2())
    class TestMatmul(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestMatmul
    ):
        pass

    class TestArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions
    ):

        @pytest.mark.parametrize("array_2", _cartesian2d_arrays_2())
        class TestAsArrayLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestAsArrayLikeFunctions
        ):
            pass

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
                (na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0) | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0),
                na.Cartesian2dVectorArray(
                    x=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0) | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0),
                    y=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) <= 0) | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) <= 0),
                )
            ]
        )
        class TestReductionFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestReductionFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames='q',
            argvalues=[
                .25,
                25 * u.percent,
                na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
                na.Cartesian2dVectorArray(x=25 * u.percent, y=35 * u.percent),
                na.Cartesian2dVectorArray(
                    x=na.ScalarLinearSpace(.25, .50, axis='q', num=3, endpoint=True),
                    y=na.ScalarLinearSpace(50 * u.percent, 75 * u.percent, axis='q', num=3, endpoint=True)
                )
            ]
        )
        class TestPercentileLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.
            TestPercentileLikeFunctions,
        ):
            pass


@pytest.mark.parametrize('array', _cartesian2d_arrays())
class TestCartesian2dVectorArray(
    AbstractTestAbstractCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArray,
):

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(x=0, y=0),
            dict(y=slice(None)),
            dict(y=na.ScalarArrayRange(0, _num_y, axis='y')),
            dict(x=na.ScalarArrayRange(0, _num_x, axis='x'), y=na.ScalarArrayRange(0, _num_y, axis='y')),
            dict(
                y=na.Cartesian2dVectorArray(
                    x=na.ScalarArrayRange(0, _num_y, axis='y'),
                    y=na.ScalarArrayRange(0, _num_y, axis='y'),
                )
            ),
            na.ScalarArray.ones(shape=dict(y=_num_y), dtype=bool),
            np.ones_like(na.Cartesian2dVectorArray(), dtype=bool, shape=dict(y=_num_y)),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            0,
            na.ScalarUniformRandomSample(-5, 5, dict(y=_num_y)),
            na.Cartesian2dVectorUniformRandomSample(-5, 5, shape_random=dict(y=_num_y)),
        ]
    )
    def test__setitem__(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
            value: float | na.ScalarArray
    ):
        super().test__setitem__(array=array, item=item, value=value)


@pytest.mark.parametrize("type_array", [na.Cartesian2dVectorArray])
class TestCartesian2dVectorArrayCreation(
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArrayCreation,
):
    @pytest.mark.parametrize("like", [None] + _cartesian2d_arrays())
    class TestFromScalarArray(
        test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArrayCreation.TestFromScalarArray,
    ):
        pass


class AbstractTestAbstractImplicitCartesian2dVectorArray(
    test_vectors_cartesian.AbstractTestAbstractImplicitCartesianVectorArray,
):
    pass


class AbstractTestAbstractCartesian2dVectorRandomSample(
    AbstractTestAbstractImplicitCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorRandomSample,
):
    pass


def _cartesian_2d_uniform_random_samples() -> list[na.Cartesian2dVectorUniformRandomSample]:
    starts = [
        0,
        na.Cartesian2dVectorArray(
            x=na.ScalarLinearSpace(0, 1, axis='x', num=_num_x),
            y=na.ScalarLinearSpace(1, 2, axis='x', num=_num_x)
        )
    ]
    stops = [
        10,
        na.UniformUncertainScalarArray(10, width=1, num_distribution=_num_distribution),
        na.Cartesian2dVectorArray(x=10, y=11),
        na.Cartesian2dVectorArray(
            x=na.ScalarLinearSpace(10, 11, axis='x', num=_num_x),
            y=na.ScalarLinearSpace(11, 12, axis='x', num=_num_x)
        ),
    ]
    units = [None, u.mm]
    shapes_random = [dict(y=_num_y)]
    return [
        na.Cartesian2dVectorUniformRandomSample(
            start=start << unit if unit is not None else start,
            stop=stop << unit if unit is not None else stop,
            shape_random=shape_random,
        ) for start in starts for stop in stops for unit in units for shape_random in shapes_random
    ]


@pytest.mark.parametrize('array', _cartesian_2d_uniform_random_samples())
class TestCartesian2dVectorUniformRandomSample(
    AbstractTestAbstractCartesian2dVectorRandomSample,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorUniformRandomSample,
):
    pass


def _cartesian_2d_normal_random_samples() -> list[na.Cartesian2dVectorNormalRandomSample]:
    centers = [
        0,
        na.Cartesian2dVectorArray(
            x=na.ScalarLinearSpace(0, 1, axis='x', num=_num_x),
            y=na.ScalarLinearSpace(1, 2, axis='x', num=_num_x)
        )
    ]
    widths = [
        10,
        na.UniformUncertainScalarArray(10, width=1, num_distribution=_num_distribution),
        na.Cartesian2dVectorArray(x=10, y=11),
        na.Cartesian2dVectorArray(
            x=na.ScalarLinearSpace(10, 11, axis='x', num=_num_x),
            y=na.ScalarLinearSpace(11, 12, axis='x', num=_num_x)
        ),
    ]
    units = [None, u.mm]
    shapes_random = [dict(y=_num_y)]
    return [
        na.Cartesian2dVectorNormalRandomSample(
            center=center << unit if unit is not None else center,
            width=width << unit if unit is not None else width,
            shape_random=shape_random,
        ) for center in centers for width in widths for unit in units for shape_random in shapes_random
    ]


@pytest.mark.parametrize('array', _cartesian_2d_normal_random_samples())
class TestCartesian2dVectorNormalRandomSample(
    AbstractTestAbstractCartesian2dVectorRandomSample,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorNormalRandomSample,
):
    pass


class AbstractTestAbstractParameterizedCartesian2dVectorArray(
    AbstractTestAbstractImplicitCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractParameterizedCartesianVectorArray,
):
    pass



def _cartesian_2d_vector_array_ranges() -> tuple[na.Cartesian2dVectorArrayRange, ...]:
    starts = (
        0,
        na.Cartesian2dVectorArray(0, 1),
    )
    stops = (
        5,
        na.Cartesian2dVectorArray(5, 5)
    )
    axes = (
        na.Cartesian2dVectorArray('x', 'y'),
    )
    _num = na.Cartesian2dVectorArray(_num_x, _num_y)
    return tuple(
        na.Cartesian2dVectorArrayRange(
            start=start,
            stop=stop,
            axis=axis,
            step=(stop - start) / _num,
        )
        for start in starts
        for stop in stops
        for axis in axes
    )


@pytest.mark.parametrize("array", _cartesian_2d_vector_array_ranges())
class TestCartesian2dVectorArrayRange(
    AbstractTestAbstractParameterizedCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArrayRange,
):
    pass


class AbstractTestAbstractCartesian2dVectorSpace(
    AbstractTestAbstractParameterizedCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorSpace,
):
    pass


def _cartesian_2d_vector_linear_spaces() -> tuple[na.Cartesian2dVectorLinearSpace, ...]:
    starts = (
        0,
        na.Cartesian2dVectorArray(0, 1),
    )
    stops = (
        5,
        na.Cartesian2dVectorArray(5, 5)
    )
    axes = (
        na.Cartesian2dVectorArray('x', 'y'),
    )
    nums = (
        na.Cartesian2dVectorArray(_num_x, _num_y),
    )
    return tuple(
        na.Cartesian2dVectorLinearSpace(
            start=start,
            stop=stop,
            axis=axis,
            num=num,
        )
        for start in starts
        for stop in stops
        for axis in axes
        for num in nums
    )


@pytest.mark.parametrize("array", _cartesian_2d_vector_linear_spaces())
class TestCartesian2dVectorLinearSpace(
    AbstractTestAbstractCartesian2dVectorSpace,
    AbstractTestAbstractCartesian2dVectorArray,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorLinearSpace,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray,
    named_arrays._vectors.tests.test_vectors.AbstractTestAbstractVectorArray,
    named_arrays.tests.test_core.AbstractTestAbstractArray,
):
    pass



from typing import Callable, TypeVar
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
from . import scalars

__all__ = [
    "ASARRAY_LIKE_FUNCTIONS",
    "RANDOM_FUNCTIONS",
    "PLT_PLOT_LIKE_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "random",
    "jacobian",
]

ASARRAY_LIKE_FUNCTIONS = (
    na.asarray,
    na.asanyarray,
)
RANDOM_FUNCTIONS = (
    na.random.uniform,
    na.random.normal,
    na.random.poisson,
)
PLT_PLOT_LIKE_FUNCTIONS = (
    na.plt.plot,
    na.plt.fill,
)
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def asarray_like(
        func: Callable,
        a: None | float | u.Quantity | na.AbstractScalarArray,
        dtype: None | type | np.dtype = None,
        order: None | str = None,
        *,
        like: None | na.AbstractScalarArray = None,
) -> None | na.ScalarArray:

    func_numpy = getattr(np, func.__name__)

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalarArray):
            a_ndarray = a.ndarray
            a_axes = a.axes
        else:
            return NotImplemented
    else:
        a_ndarray = a
        a_axes = None

    if isinstance(like, na.AbstractArray):
        if isinstance(like, na.AbstractScalarArray):
            like_ndarray = like.ndarray
            type_like = like.type_explicit
        else:
            return NotImplemented
    else:
        like_ndarray = like
        type_like = na.ScalarArray

    if isinstance(like_ndarray, u.Quantity):
        like_ndarray = like_ndarray.value
    like_ndarray = func_numpy(like_ndarray)

    return type_like(
        ndarray=func_numpy(
            a=a_ndarray,
            dtype=dtype,
            order=order,
            like=like_ndarray,
        ),
        axes=a_axes,
    )


@_implements(na.arange)
def arange(
        start: float | complex | u.Quantity | na.AbstractArray,
        stop: float | complex | u.Quantity | na.AbstractArray,
        axis: str | na.AbstractArray,
        step: int | na.AbstractArray = 1,
) -> na.ScalarArray:

    start = scalars._normalize(start)
    stop = scalars._normalize(stop)

    if start.size > 1:
        raise ValueError(f"`start` must have only one element, got shape {start.shape}")

    if stop.size > 1:
        raise ValueError(f"`stop` must have only one element, got shape {stop.shape}")

    return na.ScalarArray(
        ndarray=np.arange(
            start=start.ndarray,
            stop=stop.ndarray,
            step=step,
        ),
        axes=(axis,),
    )


@_implements(na.unit)
def unit(
        a: na.AbstractScalarArray,
        squeeze: bool = True,
) -> None | u.UnitBase:
    return na.unit(a.ndarray)


@_implements(na.unit_normalized)
def unit_normalized(
        a: na.AbstractScalarArray,
        squeeze: bool = True,
) -> u.UnitBase:
    result = na.unit(a)
    if result is None:
        result = u.dimensionless_unscaled
    return result


@_implements(na.interp)
def interp(
        x: float | u.Quantity | na.AbstractScalarArray,
        xp:  na.AbstractScalarArray,
        fp: na.AbstractScalarArray,
        axis: None | str = None,
        left: None | float | u.Quantity | na.AbstractScalarArray = None,
        right: None | float | u.Quantity | na.AbstractScalarArray = None,
        period: None | float | u.Quantity | na.AbstractScalarArray = None,
):
    try:
        x = scalars._normalize(x)
        xp = scalars._normalize(xp)
        fp = scalars._normalize(fp)
        left = scalars._normalize(left) if left is not None else left
        right = scalars._normalize(right) if right is not None else right
        period = scalars._normalize(period) if period is not None else period
    except na.ScalarTypeError:
        return NotImplemented

    if axis is None:
        if xp.ndim != 1:
            raise ValueError("if `axis` is `None`, `xp` must have only one axis")
        axis = next(iter(xp.shape))

    shape = na.shape_broadcasted(
        # x,
        xp[{axis: 0}],
        fp[{axis: 0}],
        left,
        right,
        period,
    )

    x = na.broadcast_to(x, na.broadcast_shapes(x.shape, shape))
    xp = na.broadcast_to(xp, na.broadcast_shapes(xp.shape, shape))
    fp = na.broadcast_to(fp, na.broadcast_shapes(fp.shape, shape))
    left = na.broadcast_to(left, shape) if left is not None else left
    right = na.broadcast_to(right, shape) if right is not None else right
    period = na.broadcast_to(period, shape) if period is not None else period

    result = np.empty_like(x.value)
    if fp.unit is not None:
        result = result << fp.unit

    for index in na.ndindex(shape):
        x_index = x[index]
        xp_index = xp[index]
        fp_index = fp[index]
        left_index = left[index].ndarray if left is not None else left
        right_index = right[index].ndarray if right is not None else right
        period_index = period[index].ndarray if period is not None else period

        shape_index = na.shape_broadcasted(xp_index, fp_index)

        result[index] = result.type_explicit(
            ndarray=np.interp(
                x=x_index.ndarray,
                xp=xp_index.ndarray_aligned(shape_index),
                fp=fp_index.ndarray_aligned(shape_index),
                left=left_index,
                right=right_index,
                period=period_index,
            ),
            axes=x.axes,
        )

    return result


def random(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalarArray,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity | na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if shape_random is None:
        shape_random = dict()

    shape_base = na.shape_broadcasted(*args, *kwargs.values())
    shape = na.broadcast_shapes(shape_base, shape_random)

    args = tuple(arg.ndarray_aligned(shape) for arg in args)
    kwargs = {k: kwargs[k].ndarray_aligned(shape) for k in kwargs}

    unit = None
    for a in args + tuple(kwargs.values()):
        if isinstance(a, u.Quantity):
            unit = a.unit
            break

    if unit is not None:
        args = tuple(
            arg.value if isinstance(arg, u.Quantity)
            else (arg << u.dimensionless_unscaled).to_value(unit)
            for arg in args
        )
        kwargs = {
            k: kwargs[k].value if isinstance(kwargs[k], u.Quantity)
            else (kwargs[k] << u.dimensionless_unscaled).to_value(unit)
            for k in kwargs
        }

    if seed is None:
        func = getattr(np.random, func.__name__)
    else:
        func = getattr(np.random.default_rng(seed), func.__name__)

    value = func(
        *args,
        size=tuple(shape.values()),
        **kwargs,
    )

    if unit is not None:
        value = value << unit

    return na.ScalarArray(
        ndarray=value,
        axes=tuple(shape.keys()),
    )


def plt_plot_like(
        func: Callable,
        *args: na.AbstractScalarArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:

    if components is not None:
        raise ValueError(f"`components` should be `None` for scalars, got {components}")

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        where = scalars._normalize(where)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    shape = na.shape_broadcasted(*args)

    if axis is None:
        if len(shape) != 1:
            raise ValueError(
                f"if `axis` is `None`, the broadcasted shape of `*args`, {shape}, should have one element"
            )
        axis = next(iter(shape))

    shape_orthogonal = {a: shape[a] for a in shape if a != axis}

    args = tuple(arg.broadcast_to(shape) for arg in args)

    if not set(ax.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `ax`, {ax.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}",
        )
    ax = ax.broadcast_to(shape_orthogonal)

    if not set(where.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `where`, {where.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
        )
    where = where.broadcast_to(shape_orthogonal)

    kwargs_broadcasted = dict()
    for k in kwargs:
        kwarg = kwargs[k]
        if not set(na.shape(kwarg)).issubset(shape_orthogonal):
            raise ValueError(
                f"the shape of `{k}`, {na.shape(kwarg)}, "
                f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
            )
        kwargs_broadcasted[k] = na.broadcast_to(kwarg, shape_orthogonal)
    kwargs = kwargs_broadcasted

    result = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        if where[index]:
            func_matplotlib = getattr(ax[index].ndarray, func.__name__)
            args_index = tuple(arg[index].ndarray for arg in args)
            kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
            result[index] = func_matplotlib(
                *args_index,
                **kwargs_index,
            )[0]

    return result


@_implements(na.jacobian)
def jacobian(
        function: Callable[[na.AbstractScalar], na.AbstractScalar],
        x: na.AbstractScalar,
        dx: None | na.AbstractScalar = None,
        like: None | na.AbstractScalar = None,
) -> na.AbstractScalar:

    f = function(x)

    if isinstance(f, na.AbstractScalar):
        if isinstance(x, na.AbstractScalar):
            x0 = x + dx
            f0 = function(x0)
            df = f0 - f
            return df / dx

        else:
            return NotImplemented
    else:
        return NotImplemented


@_implements(na.optimize.root_newton)
def optimize_root_newton(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        jacobian: Callable[[na.ScalarLike], na.ScalarLike],
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.ScalarArray:

    if isinstance(guess, na.AbstractArray):
        if isinstance(guess, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        guess = na.ScalarArray(guess)

    x = guess
    f = function(x)

    if isinstance(f, na.AbstractArray):
        if isinstance(f, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        f = na.ScalarArray(f)

    if isinstance(max_abs_error, na.AbstractArray):
        if isinstance(max_abs_error, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        max_abs_error = na.ScalarArray(max_abs_error)

    if na.shape(max_abs_error):
        raise ValueError(f"argument `max_abs_error` should have an empty shape, got {na.shape(max_abs_error)}")

    shape = na.shape_broadcasted(f, guess)

    converged = na.broadcast_to(0 * na.value(f), shape=shape).astype(bool)

    x = na.broadcast_to(x, shape).astype(float)

    for i in range(max_iterations):

        if callback is not None:
            callback(i, x, f, converged)

        converged |= np.abs(f) < max_abs_error

        if np.all(converged):
            return x

        jac = jacobian(x)

        correction = f / jac

        x = x - correction

        f = function(x)

    raise ValueError("Max iterations exceeded")


@_implements(na.optimize.root_secant)
def optimize_root_secant(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        min_step_size: na.ScalarLike,
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        damping: None | float = None,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.ScalarArray:

    try:
        guess = scalars._normalize(guess)

        min_step_size = scalars._normalize(min_step_size)

        x0 = guess - 10 * min_step_size
        x1 = guess

        f0 = function(x0)
        f0 = scalars._normalize(f0)

        max_abs_error = scalars._normalize(max_abs_error)

    except scalars.ScalarTypeError:
        return NotImplemented

    if na.shape(max_abs_error):
        raise ValueError(f"argument `max_abs_error` should have an empty shape, got {na.shape(max_abs_error)}")

    shape = na.shape_broadcasted(f0, guess, min_step_size)

    converged = na.broadcast_to(0 * na.value(f0), shape=shape).astype(bool)

    x1 = na.broadcast_to(x1, shape).astype(float)

    for i in range(max_iterations):


        f1 = function(x1)

        if callback is not None:
            callback(i, x1, f1, converged)

        if max_abs_error is not None:
            converged |= np.abs(f1) < max_abs_error

        dx = x1 - x0

        converged |= np.abs(dx) < np.abs(min_step_size)

        if np.all(converged):
            return x1

        active = ~converged

        df = f1 - f0

        if np.any(df == 0, where=active):
            raise ValueError("stationary point detected")

        dx_active = dx[active]
        f0_active = f0[active]
        f1_active = f1[active]

        df_active = f1_active - f0_active
        if np.any(df_active == 0):
            raise ValueError("stationary point detected")

        jacobian = df_active / dx_active

        correction = f1_active / jacobian
        if damping is not None:
            correction = damping * correction

        x2 = x1.copy()
        x2[active] -= correction

        x0 = x1
        x1 = x2
        f0 = f1

    raise ValueError("Max iterations exceeded")



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import numpy as np
import named_arrays as na

__all__ = [
    'AbstractCartesian2dMatrixArray',
    'Cartesian2dMatrixArray',
    'AbstractImplicitCartesian2dMatrixArray',
    'Cartesian2dIdentityMatrixArray',
    'AbstractCartesian2dRotationMatrixArray',
    'Cartesian2dRotationMatrixArray'
]

XT = TypeVar('XT', bound=na.AbstractVectorArray)
YT = TypeVar('YT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dMatrixArray(
    na.AbstractCartesian2dVectorArray,
    na.AbstractCartesianMatrixArray,
):
    @property
    @abc.abstractmethod
    def x(self) -> na.AbstractVectorArray:
        """
        The `x` row of the matrix.
        """

    @property
    @abc.abstractmethod
    def y(self) -> na.AbstractVectorArray:
        """
        The `y` row of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractCartesian2dMatrixArray]:
        return AbstractCartesian2dMatrixArray

    @property
    def type_explicit(self) -> Type[Cartesian2dMatrixArray]:
        return Cartesian2dMatrixArray



    @property
    def type_vector(self) -> Type[na.Cartesian2dVectorArray]:
        return na.Cartesian2dVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        if not self.is_square:
            raise ValueError("can only compute determinant of square matrices")
        xx, xy = self.x.components.values()
        yx, yy = self.y.components.values()
        return xx * yy - xy * yx


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dMatrixArray(
    na.Cartesian2dVectorArray,
    AbstractCartesian2dMatrixArray,
    na.AbstractExplicitMatrixArray,
    Generic[XT, YT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian2dMatrixArray(
    na.AbstractImplicitCartesian2dVectorArray,
    AbstractCartesian2dMatrixArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dIdentityMatrixArray(
    AbstractImplicitCartesian2dMatrixArray
):

    @property
    def explicit(self):
        return na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=1, y=0),
            y=na.Cartesian2dVectorArray(x=0, y=1),
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian2dRotationMatrixArray(
    AbstractImplicitCartesian2dMatrixArray,
):

    @property
    @abc.abstractmethod
    def angle(self) -> na.ScalarLike:
        """
        The angle of the rotation matrix.
        """


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian2dRotationMatrixArray(
    AbstractCartesian2dRotationMatrixArray,
):
    angle: na.ScalarLike = dataclasses.MISSING

    @property
    def explicit(self) -> na.Cartesian2dMatrixArray:
        a = self.angle
        return na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=np.cos(a), y=-np.sin(a)),
            y=na.Cartesian2dVectorArray(x=np.sin(a), y=np.cos(a)),
        )



from __future__ import annotations
from typing import Type, Generic, TypeVar
from typing_extensions import Self
import abc
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractPositionalVectorArray',
    'PositionalVectorArray',
    'AbstractImplicitPositionalVectorArray',
    'AbstractParameterizedPositionalVectorArray',
    'AbstractPositionalVectorSpace',
    'PositionalVectorLinearSpace',

]

PositionT = TypeVar("PositionT", bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionalVectorArray(
    na.AbstractCartesianVectorArray,
):

    @property
    @abc.abstractmethod
    def position(self) -> na.ArrayLike:
        """
        The `position` component of the vector.
        """

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractPositionalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return PositionalVectorArray

    @property
    def type_matrix(self) -> Type[na.PositionalMatrixArray]:
        return na.PositionalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class PositionalVectorArray(
    AbstractPositionalVectorArray,
    na.AbstractExplicitCartesianVectorArray,
    Generic[PositionT],
):
    position: PositionT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> PositionalVectorArray:
        result = super().from_scalar(scalar, like=like)
        if result is not NotImplemented:
            return result
        return cls(position=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitPositionalVectorArray(
    AbstractPositionalVectorArray,
    na.AbstractImplicitCartesianVectorArray,
):

    @property
    def position(self) -> na.ArrayLike:
        return self.explicit.position


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedPositionalVectorArray(
    AbstractImplicitPositionalVectorArray,
    na.AbstractParameterizedVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionalVectorSpace(
    AbstractParameterizedPositionalVectorArray,
    na.AbstractVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class PositionalVectorLinearSpace(
    AbstractPositionalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass



from __future__ import annotations
from typing import TypeVar, Generic, Type
import abc
import dataclasses
import named_arrays as na
import numpy as np

__all__ = [
    'AbstractPositionalMatrixArray',
    'PositionalMatrixArray',
]

PositionT = TypeVar('PositionT', bound=na.AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPositionalMatrixArray(
    na.AbstractCartesianMatrixArray,
    na.AbstractPositionalVectorArray,
):
    @property
    @abc.abstractmethod
    def position(self) -> na.AbstractVectorArray:
        """
        The `position` component of the matrix.
        """

    @property
    def type_abstract(self) -> Type[AbstractPositionalMatrixArray]:
        return AbstractPositionalMatrixArray

    @property
    def type_explicit(self) -> Type[PositionalMatrixArray]:
        return PositionalMatrixArray

    @property
    def type_vector(self) -> Type[na.PositionalVectorArray]:
        return na.PositionalVectorArray

    @property
    def determinant(self) -> na.ScalarLike:
        raise NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class PositionalMatrixArray(
    na.PositionalVectorArray,
    AbstractPositionalMatrixArray,
    na.AbstractExplicitMatrixArray,
    Generic[PositionT],
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
from . import test_vectors
from ..cartesian.tests import test_vectors_cartesian

_num_x = test_vectors._num_x
_num_y = test_vectors._num_y
_num_z = test_vectors._num_z
_num_distribution = test_vectors._num_distribution


def _spectral_arrays() -> list[na.SpectralVectorArray]:
    return [
        na.SpectralVectorArray(wavelength=500 * u.nm),
        na.SpectralVectorArray(wavelength=na.linspace(400, 600, axis="y", num=_num_y) * u.nm),
    ]


def _spectral_arrays_2() -> list[na.SpectralVectorArray]:
    return [
        na.SpectralVectorArray(wavelength=400 * u.nm),
        na.SpectralVectorArray(wavelength=na.NormalUncertainScalarArray(400 * u.nm, width=1 * u.nm))
    ]


def _spectral_items() -> list[na.AbstractSpectralVectorArray | dict[str, int | slice | na.AbstractSpectralVectorArray]]:
    return [
        dict(y=0),
        dict(y=slice(0, 1)),
        dict(y=na.ScalarArrayRange(0, 2, axis='y')),
    ]


class AbstractTestAbstractSpectralVectorArray(
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray,
):
    def test_wavelength(self, array: na.AbstractSpectralVectorArray):
        assert isinstance(na.as_named_array(array.wavelength), na.AbstractScalar)
        assert np.all(array.wavelength > 0)

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=_spectral_items()
    )
    def test__getitem__(
            self,
            array: na.AbstractSpectralVectorArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize('array_2', _spectral_arrays_2())
    class TestUfuncBinary(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize('array_2', _spectral_arrays_2())
    class TestMatmul(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestMatmul
    ):
        pass

    class TestArrayFunctions(
        test_vectors.AbstractTestAbstractVectorArray.TestArrayFunctions
    ):

        @pytest.mark.parametrize("array_2", _spectral_arrays_2())
        class TestAsArrayLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestAsArrayLikeFunctions
        ):
            pass

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
            ]
        )
        class TestReductionFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestReductionFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames='q',
            argvalues=[
                .25,
                25 * u.percent,
                na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
            ]
        )
        class TestPercentileLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions
            .TestPercentileLikeFunctions,
        ):
            pass

    class TestNamedArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions,
    ):
        @pytest.mark.xfail(raises=TypeError)
        class TestPltPlotLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestPltPlotLikeFunctions,
        ):
            pass


@pytest.mark.parametrize("array", _spectral_arrays())
class TestSpectralVectorArray(
    AbstractTestAbstractSpectralVectorArray,
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArray,
):
    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(y=slice(None)),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            700 * u.nm,
        ]
    )
    def test__setitem__(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
            value: float | na.ScalarArray
    ):
        super().test__setitem__(array=array, item=item, value=value)


class AbstractTestAbstractImplicitSpectralVectorArray(
    AbstractTestAbstractSpectralVectorArray,
    test_vectors_cartesian.AbstractTestAbstractImplicitCartesianVectorArray,
):
    pass


class AbstractTestAbstractParameterizedSpectralVectorArray(
    AbstractTestAbstractImplicitSpectralVectorArray,
    test_vectors_cartesian.AbstractTestAbstractParameterizedCartesianVectorArray,
):
    pass


class AbstractTestAbstractSpectralVectorSpace(
    AbstractTestAbstractParameterizedSpectralVectorArray,
    test_vectors_cartesian.AbstractTestAbstractParameterizedCartesianVectorArray,
):
    pass


def _spectral_linear_spaces() -> list[na.SpectralVectorLinearSpace]:
    return [
        na.SpectralVectorLinearSpace(400 * u.nm, 600 * u.nm, axis="y", num=_num_y)
    ]


@pytest.mark.parametrize("array", _spectral_linear_spaces())
class TestSpectralVectorLinearSpace(
    AbstractTestAbstractSpectralVectorSpace,
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorLinearSpace,
):
    pass



import named_arrays._vectors.cartesian.tests.test_vectors_cartesian
import named_arrays._matrices.tests.test_matrices

__all__ = [
    'AbstractTestAbstractCartesianMatrixArray',
    'AbstractTestAbstractExplicitCartesianMatrixArray',
    'AbstractTestAbstractExplicitCartesianMatrixArrayCreation',
    'AbstractTestAbstractImplicitCartesianMatrixArray',
    'AbstractTestAbstractCartesianMatrixRandomSample',
    'AbstractTestAbstractCartesianMatrixUniformRandomSample',
    'AbstractTestAbstractCartesianMatrixNormalRandomSample',
    'AbstractTestAbstractParameterizedCartesianMatrixArray',
]


class AbstractTestAbstractCartesianMatrixArray(
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractMatrixArray,
    named_arrays._vectors.cartesian.tests.test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray,
):
    pass


class AbstractTestAbstractExplicitCartesianMatrixArray(
    AbstractTestAbstractCartesianMatrixArray,
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractExplicitMatrixArray,
):
    pass


class AbstractTestAbstractExplicitCartesianMatrixArrayCreation(
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractExplicitMatrixArrayCreation,
):
    pass


class AbstractTestAbstractImplicitCartesianMatrixArray(
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractImplicitMatrixArray,
):
    pass


class AbstractTestAbstractCartesianMatrixRandomSample(
    AbstractTestAbstractImplicitCartesianMatrixArray,
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractMatrixRandomSample,
):
    pass


class AbstractTestAbstractCartesianMatrixUniformRandomSample(
    AbstractTestAbstractCartesianMatrixRandomSample,
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractMatrixUniformRandomSample,
):
    pass


class AbstractTestAbstractCartesianMatrixNormalRandomSample(
    AbstractTestAbstractCartesianMatrixRandomSample,
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractMatrixNormalRandomSample,
):
    pass


class AbstractTestAbstractParameterizedCartesianMatrixArray(
    AbstractTestAbstractImplicitCartesianMatrixArray,
    named_arrays._matrices.tests.test_matrices.AbstractTestAbstractParameterizedMatrixArray,
):
    pass



from __future__ import annotations
from typing import ClassVar, Type, Sequence, Callable, Collection, Any, Generic, TypeVar
from typing_extensions import Self
import abc
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    'VectorPrototypeT',
    'VectorTypeError',
    'AbstractVectorArray',
    'AbstractScalarOrVectorArray',
    'AbstractExplicitVectorArray',
    'AbstractImplicitVectorArray',
    'AbstractVectorRandomSample',
    'AbstractVectorUniformRandomSample',
    'AbstractVectorNormalRandomSample',
    'AbstractParameterizedVectorArray',
    'AbstractVectorArrayRange',
    'AbstractVectorSpace',
    'AbstractVectorLinearSpace',
    'AbstractVectorStratifiedRandomSpace',
    'AbstractVectorLogarithmicSpace',
    'AbstractVectorGeometricSpace',
]

VectorPrototypeT = TypeVar("VectorPrototypeT", bound="AbstractVectorArray")
VectorStartT = TypeVar('VectorStartT', bound="float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray")
VectorStopT = TypeVar('VectorStopT', bound="float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray")


class VectorTypeError(TypeError):
    pass


def _prototype(*arrays: float | u.Quantity | na.AbstractArray) -> na.AbstractVectorArray:
    for array in arrays:
        if isinstance(array, na.AbstractVectorArray):
            return array

    raise VectorTypeError


def _normalize(
        a: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
        prototype: VectorPrototypeT,
) -> VectorPrototypeT:
    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractVectorArray):
            if a.type_abstract == prototype.type_abstract:
                result = a
            else:
                raise VectorTypeError
        elif isinstance(a, na.AbstractScalar):
            result = prototype.type_explicit.from_scalar(a)
        else:
            raise VectorTypeError
    else:
        result = prototype.type_explicit.from_scalar(a)

    return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorArray(
    na.AbstractArray
):
    __named_array_priority__: ClassVar[float] = 100 * na.AbstractScalarArray.__named_array_priority__

    @property
    @abc.abstractmethod
    def type_explicit(self: Self) -> Type[AbstractExplicitVectorArray]:
        pass

    @property
    @abc.abstractmethod
    def type_matrix(self) -> Type[na.AbstractExplicitMatrixArray]:
        """
        The corresponding :class:`named_arrays.AbstractMatrixArray` class
        """

    @property
    def matrix(self) -> na.AbstractMatrixArray:
        new_dict = {}
        for c in self.components:
            component = self.components[c]
            if isinstance(component, AbstractVectorArray):
                new_dict[c] = component.matrix
            elif isinstance(na.as_named_array(component), na.AbstractScalar):
                new_dict[c] = component
            else:
                raise NotImplementedError


        return self.type_matrix.from_components(new_dict)

    @property
    def cartesian_nd(self) -> na.AbstractCartesianNdVectorArray:
        """
        Convert any instance of :class:`AbstractVectorArray` to an instance of :class:`AbstractCartesianNdVectorArray`
        """
        components_new = dict()
        components = self.components
        for c in components:

            component = components[c]
            if isinstance(component, na.AbstractVectorArray):
                component2 = component.cartesian_nd.components
                for c2 in component2:
                    components_new[f"{c}_{c2}"] = component2[c2]
            else:
                components_new[c] = component

        return na.CartesianNdVectorArray(components_new)

    @property
    @abc.abstractmethod
    def components(self: Self) -> dict[str, na.ArrayLike]:
        """
        The vector components of this array expressed as a :class:`dict` where the keys are the names of the component.
        """
        return dict()

    @property
    def entries(self) -> dict[str, na.ArrayLike]:
        """
        The scalar entries that compose this object.
        """
        return self.cartesian_nd.components

    @property
    def value(self) -> na.AbstractExplicitVectorArray:
        components = self.components
        components = {c: na.value(components[c]) for c in components}
        return self.type_explicit.from_components(components)

    @property
    def centers(self: Self) -> AbstractExplicitVectorArray:
        components = self.components
        components_result = dict()
        for c in components:
            if isinstance(components[c], na.AbstractArray):
                components_result[c] = components[c].centers
            else:
                components_result[c] = components[c]
        return self.type_explicit.from_components(components_result)

    @property
    def prototype_vector(self) -> na.AbstractExplicitVectorArray:
        """
        Return vector of same type with all components zeroed.
        """
        return self.type_explicit.from_components(dict.fromkeys(self.components, 0))

    def astype(
            self: Self,
            dtype: str | np.dtype | Type | dict[str, str | np.dtype | Type],
            order: str = 'K',
            casting='unsafe',
            subok: bool = True,
            copy: bool = True,
    ) -> Self:
        components = self.components

        components_normalized = dict()
        for c in components:
            if not isinstance(components[c], (np.ndarray, na.AbstractArray)):
                components_normalized[c] = np.array(components[c])
            else:
                components_normalized[c] = components[c]
        components = components_normalized

        if not isinstance(dtype, dict):
            dtype = {c: dtype for c in components}
        kwargs = dict(
            order=order,
            casting=casting,
            subok=subok,
            copy=copy,
        )
        return self.type_explicit.from_components(
            {c: components[c].astype(dtype=dtype[c], **kwargs) for c in components})

    def to(
        self: Self,
        unit: u.UnitBase | dict[str, None | u.UnitBase],
        equivalencies: None | list[tuple[u.Unit, u.Unit]] = [],
        copy: bool = True,
    ) -> AbstractExplicitVectorArray:
        components = self.components
        if not isinstance(unit, dict):
            unit = {c: unit for c in components}
        components_result = dict()
        for c in components:
            if unit[c] is not None:
                if not isinstance(components[c], (u.Quantity, na.AbstractArray)):
                    components_c = components[c] << u.dimensionless_unscaled
                else:
                    components_c = components[c]
                components_result[c] = components_c.to(
                    unit=unit[c],
                    equivalencies=equivalencies,
                    copy=copy,
                )
            else:
                components_result[c] = components[c]
        return self.type_explicit.from_components(components_result)

    def add_axes(self: Self, axes: str | Sequence[str]) -> AbstractExplicitVectorArray:
        components = self.components
        return self.type_explicit.from_components({c: na.add_axes(components[c], axes) for c in components})

    def combine_axes(
            self: Self,
            axes: Sequence[str] = None,
            axis_new: str = None,
    ) -> AbstractExplicitVectorArray:

        shape = self.shape

        if axes is None:
            axes = tuple(shape)

        shape_base = {ax: shape[ax] for ax in shape if ax in axes}

        components = self.components
        components_result = dict()
        for c in components:
            shape_c = na.shape(components[c]) | shape_base
            components_result[c] = na.broadcast_to(components[c], shape=shape_c).combine_axes(
                axes=axes,
                axis_new=axis_new
            )

        return self.type_explicit.from_components(components_result)

    def _getitem(
            self: Self,
            item: dict[str, int | slice | AbstractScalarOrVectorArray] | AbstractScalarOrVectorArray,
    ) -> Self:

        array = self.explicit
        shape_array = array.shape
        components = array.components

        if isinstance(item, na.AbstractArray):
            item = item.explicit
            shape_item = item.shape

            shape_base = {ax: shape_array[ax] for ax in shape_item if ax in shape_array}
            for c in components:
                component = na.as_named_array(components[c])
                components[c] = component.broadcast_to(na.broadcast_shapes(component.shape, shape_base))

            if item.type_abstract == self.type_abstract:
                item_accumulated = True
                components_item = item.components
                for c in components_item:
                    item_accumulated = item_accumulated & components_item[c]
                item = self.type_explicit.from_scalar(item_accumulated, like=self)
            elif isinstance(item, na.AbstractScalar):
                item = self.type_explicit.from_scalar(item, like=self)
            else:
                return NotImplemented

        elif isinstance(item, dict):
            shape_base = {ax: shape_array[ax] for ax in item if ax in shape_array}
            for c in components:
                component = na.as_named_array(components[c])
                components[c] = component.broadcast_to(na.broadcast_shapes(component.shape, shape_base))

            item = item.copy()
            for ax in item:
                if isinstance(item[ax], na.AbstractArray):
                    if item[ax].type_abstract == self.type_abstract:
                        item[ax] = item[ax].explicit
                    elif isinstance(item[ax], na.AbstractScalar):
                        item[ax] = self.type_explicit.from_scalar(item[ax], like=self)
                    else:
                        return NotImplemented
                elif isinstance(item[ax], (int, slice)):
                    item[ax] = self.type_explicit.from_scalar(item[ax], like=self)
                elif item[ax] is None:
                    item[ax] = self.type_explicit.from_scalar(item[ax], like=self)
                else:
                    return NotImplemented

        else:
            return NotImplemented

        components_result = dict()
        for c in components:
            if isinstance(item, dict):
                components_result[c] = na.as_named_array(components[c])[{ax: item[ax].components[c] for ax in item}]
            else:
                components_result[c] = components[c][na.as_named_array(item.components[c])]

        return self.type_explicit.from_components(components_result)

    def _getitem_reversed(
            self: Self,
            array: na.ScalarArray,
            item: dict[str, int | slice | AbstractVectorArray] | AbstractVectorArray,
    ):
        if array.type_abstract == self.type_abstract:
            pass
        elif isinstance(array, na.AbstractArray):
            array = self.type_explicit.from_scalar(array, like=self)
        else:
            return NotImplemented
        return array._getitem(item)

    def __bool__(self: Self) -> bool:
        result = super().__bool__()
        components = self.components
        for c in components:
            component = bool(components[c])
            result = result and component
        return result

    def __array_matmul__(
            self: Self,
            x1: na.ArrayLike,
            x2: na.ArrayLike,
            out: tuple[None | na.AbstractExplicitArray] = (None,),
            **kwargs,
    ) -> na.AbstractExplicitArray:

        result = super().__array_matmul__(
            x1=x1,
            x2=x2,
            out=out,
            **kwargs,
        )
        if result is not NotImplemented:
            return result

        out = out[0]

        if isinstance(x1, AbstractVectorArray):
            if isinstance(x2, na.AbstractVectorArray):
                components_x1 = x1.cartesian_nd.broadcasted.components
                components_x2 = x2.cartesian_nd.broadcasted.components

                if components_x1.keys() == components_x2.keys():
                    result = 0
                    for c in components_x1:
                        component_x1 = na.as_named_array(components_x1[c])
                        component_x2 = na.as_named_array(components_x2[c])
                        result = np.add(result, np.matmul(component_x1, component_x2), out=out)
                else:
                    result = NotImplemented
            elif isinstance(na.as_named_array(x2), na.ScalarArray):
                result = np.multiply(x1, x2, out=out, **kwargs)
            else:
                result = NotImplemented

        elif isinstance(na.as_named_array(x1), na.ScalarArray):
            if isinstance(x2, na.AbstractVectorArray):
                result = np.multiply(x1, x2, out=out, **kwargs)
            else:
                result = NotImplemented

        else:
            result = NotImplemented

        return result

    def __array_function__(
            self: Self,
            func: Callable,
            types: Collection,
            args: tuple,
            kwargs: dict[str, Any],
    ):
        result = super().__array_function__(func=func, types=types, args=args, kwargs=kwargs)
        if result is not NotImplemented:
            return result

        from . import vector_array_functions

        if func in vector_array_functions.SINGLE_ARG_FUNCTIONS:
            return vector_array_functions.array_functions_single_arg(func, *args, **kwargs)

        if func in vector_array_functions.ARRAY_CREATION_LIKE_FUNCTIONS:
            return vector_array_functions.array_function_array_creation_like(func, *args, **kwargs)

        if func in vector_array_functions.SEQUENCE_FUNCTIONS:
            return vector_array_functions.array_function_sequence(func, *args, **kwargs)

        if func in vector_array_functions.DEFAULT_FUNCTIONS:
            return vector_array_functions.array_function_default(func, *args, **kwargs)

        if func in vector_array_functions.PERCENTILE_LIKE_FUNCTIONS:
            return vector_array_functions.array_function_percentile_like(func, *args, **kwargs)

        if func in vector_array_functions.ARG_REDUCE_FUNCTIONS:
            return vector_array_functions.array_function_arg_reduce(func, *args, **kwargs)

        if func in vector_array_functions.FFT_LIKE_FUNCTIONS:
            return vector_array_functions.array_function_fft_like(func, *args, **kwargs)

        if func in vector_array_functions.FFTN_LIKE_FUNCTIONS:
            return vector_array_functions.array_function_fftn_like(func, *args, **kwargs)

        if func in vector_array_functions.STACK_LIKE_FUNCTIONS:
            return vector_array_functions.array_function_stack_like(func, *args, **kwargs)

        if func in vector_array_functions.HANDLED_FUNCTIONS:
            return vector_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented

    def __named_array_function__(self, func, *args, **kwargs):
        result = super().__named_array_function__(func, *args, **kwargs)
        if result is not NotImplemented:
            return result

        from . import vector_named_array_functions

        if func in vector_named_array_functions.ASARRAY_LIKE_FUNCTIONS:
            return vector_named_array_functions.asarray_like(func=func, *args, **kwargs)

        if func in vector_named_array_functions.RANDOM_FUNCTIONS:
            return vector_named_array_functions.random(func=func, *args, **kwargs)

        if func in vector_named_array_functions.PLT_PLOT_LIKE_FUNCTIONS:
            return vector_named_array_functions.plt_plot_like(func, *args, **kwargs)

        if func in vector_named_array_functions.HANDLED_FUNCTIONS:
            return vector_named_array_functions.HANDLED_FUNCTIONS[func](*args, **kwargs)

        return NotImplemented


AbstractScalarOrVectorArray = na.AbstractScalar | AbstractVectorArray


@dataclasses.dataclass(eq=False, repr=False)
class AbstractExplicitVectorArray(
    AbstractVectorArray,
    na.AbstractExplicitArray,
):
    @classmethod
    def from_scalar_array(
            cls: Type[Self],
            a: None | float | u.Quantity | na.AbstractArray,
            like: None | AbstractExplicitVectorArray = None,
    ) -> AbstractExplicitVectorArray:

        self = super().from_scalar_array(a=a, like=like)

        components_self = dict()

        if like is None:
            for c in self.components:
                components_self[c] = a
        else:
            components_like = like.components
            for c in components_like:
                component_like = components_like[c]
                if isinstance(component_like, na.AbstractArray):
                    components_self[c] = component_like.from_scalar_array(a, like=component_like)
                else:
                    components_self[c] = a

        self.components = components_self

        return self

    @classmethod
    def from_components(
            cls: Type[Self],
            components: dict[str, na.AbstractArray],
    ) -> AbstractExplicitVectorArray:
        return cls(**components)

    @classmethod
    @abc.abstractmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.ScalarLike,
            like: None | AbstractExplicitVectorArray = None,
    ) -> AbstractExplicitVectorArray:
        """
        Convert a scalar (an instance of :class:`named_arrays.AbstractScalar`) into a vector.
        """

        if like is not None:
            return like.type_explicit.from_components({c: scalar for c in like.components})
        else:
            return NotImplemented

    @classmethod
    def from_cartesian_nd(
            cls: AbstractExplicitVectorArray,
            array: na.CartesianNdVectorArray,
            like: None | AbstractExplicitVectorArray = None,
    ) -> AbstractExplicitVectorArray:

        if like is None:
            components_new = array.components

        else:
            nd_components = array.components
            components_new = {}
            components = like.components
            for c in components:

                component = components[c]
                if isinstance(component, na.AbstractVectorArray):
                    nd_key_mod = f"{c}_"
                    sub_dict = {k[len(nd_key_mod):]: v for k, v in nd_components.items() if k.startswith(nd_key_mod)}
                    components_new[c] = component.type_explicit.from_cartesian_nd(
                        na.CartesianNdMatrixArray(sub_dict),
                        like=component
                    )
                else:
                    components_new[c] = nd_components[c]

        return cls.from_components(components_new)

    @property
    def components(self: Self) -> dict[str, na.ArrayLike]:
        return self.__dict__

    @components.setter
    def components(self, value: dict[str, na.ArrayLike]) -> None:
        self.__dict__ = value

    @property
    def axes(self: Self) -> tuple[str, ...]:
        return tuple(self.shape.keys())

    @property
    def shape(self: Self) -> dict[str, int]:
        return na.shape_broadcasted(*self.components.values())

    @property
    def ndim(self: Self) -> int:
        return len(self.shape)

    @property
    def size(self: Self) -> int:
        return int(np.array(tuple(self.shape.values())).prod())

    @property
    def explicit(self: Self) -> AbstractExplicitVectorArray:
        components = self.components
        components_result = dict()
        for c in components:
            if isinstance(components[c], na.AbstractArray):
                components_result[c] = components[c].explicit
            else:
                components_result[c] = components[c]
        return self.type_explicit.from_components(components_result)

    def __setitem__(
            self,
            item: dict[str, int | slice | AbstractScalarOrVectorArray] | AbstractScalarOrVectorArray,
            value: AbstractScalarOrVectorArray,
    ):
        components_self = self.components

        if isinstance(item, na.AbstractArray):
            if isinstance(item, na.AbstractVectorArray):
                if item.type_abstract == self.type_abstract:
                    components_item = item.components
                else:
                    raise TypeError(
                        f"if `item` is an instance of `{na.AbstractVectorArray.__name__}`, "
                        f"`item.type_abstract`, `{item.type_abstract}`, "
                        f"should be equal to `self.type_abstract`, `{self.type_abstract}`"
                    )
            elif isinstance(item, na.AbstractScalar):
                components_item = {c: item for c in components_self}
            else:
                raise TypeError(
                    f"if `item` is an instance of `{na.AbstractArray.__name__}`, "
                    f"it must be an instance of `{na.AbstractVectorArray.__name__}` "
                    f"or `{na.AbstractScalar.__name__}`, got `{type(item)}`"
                )

        elif isinstance(item, dict):

            components_item = {c: dict() for c in components_self}

            for axis in item:
                item_axis = item[axis]
                if isinstance(item_axis, na.AbstractArray):
                    if isinstance(item_axis, na.AbstractVectorArray):
                        if item_axis.type_abstract == self.type_abstract:
                            components_item_axis = item_axis.components
                            for c in components_item:
                                components_item[c][axis] = components_item_axis[c]
                        else:
                            raise TypeError(
                                f"if `item['{axis}']` is an instance of `{na.AbstractVectorArray.__name__}`, "
                                f"`item['{axis}'].type_abstract`, `{item_axis.type_abstract}`, "
                                f"should be equal to `self.type_abstract`, `{self.type_abstract}`"
                            )
                    elif isinstance(item_axis, na.AbstractScalar):
                        for c in components_item:
                            components_item[c][axis] = item_axis
                    else:
                        raise TypeError(
                            f"if `item['{axis}']` is an instance of `{na.AbstractArray.__name__}`, "
                            f"it must be an instance of `{na.AbstractVectorArray.__name__}` "
                            f"or `{na.AbstractScalar.__name__}`, got `{type(item_axis)}`"
                        )
                else:
                    for c in components_item:
                        components_item[c][axis] = item_axis

        else:
            raise TypeError(
                f"`item` must be an instance of `{na.AbstractArray.__name__}` or {dict.__name__}, "
                f"got `{type(item)}`"
            )

        if isinstance(value, na.AbstractArray):
            if isinstance(value, na.AbstractVectorArray):
                if value.type_abstract == self.type_abstract:
                    components_value = value.components
                else:
                    raise TypeError(
                        f"if `value` is an instance of `{na.AbstractVectorArray.__name__}`, "
                        f"`value.type_abstract`, `{value.type_abstract}`, "
                        f"must be equal to `self.type_abstract`, `{self.type_abstract}`"
                    )
            elif isinstance(value, na.AbstractScalar):
                components_value = {c: value for c in components_self}
            else:
                raise TypeError(
                    f"if `value` is an instance of `{na.AbstractArray.__name__}`, "
                    f"it must be an instance of `{na.AbstractVectorArray.__name__}` "
                    f"or `{na.AbstractScalar.__name__}`, got `{type(value)}`"
                )
        else:
            components_value = {c: value for c in components_self}

        for c in components_self:
            components_self[c][components_item[c]] = components_value[c]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitVectorArray(
    AbstractVectorArray,
    na.AbstractImplicitArray,
):
    @property
    def components(self) -> dict[str, na.ArrayLike]:
        return self.explicit.components

    def _attr_normalized(self, name: str) -> AbstractExplicitVectorArray:

        attr = getattr(self, name)

        if isinstance(attr, na.AbstractArray):
            if attr.type_abstract == self.type_abstract:
                result = attr
            elif isinstance(attr, na.AbstractScalar):
                result = self.type_explicit.from_scalar(attr)
            else:
                raise ValueError(
                    f"if `{name}` is an instance of `AbstractArray` it must be an instance of `{self.type_abstract}` "
                    f"or `AbstractScalar`, got {type(attr)}"
                )
        else:
            result = self.type_explicit.from_scalar(attr)

        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorRandomSample(
    AbstractImplicitVectorArray,
    na.AbstractRandomSample,
):
    pass


StartT = TypeVar('StartT', bound=float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray)
StopT = TypeVar('StopT', bound=float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorUniformRandomSample(
    AbstractVectorRandomSample,
    na.AbstractUniformRandomSample[VectorStartT, VectorStopT],
):
    pass


CenterT = TypeVar('CenterT', bound=float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray)
WidthT = TypeVar('WidthT', bound=float | complex | u.Quantity | na.AbstractScalar | AbstractVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorNormalRandomSample(
    AbstractVectorRandomSample,
    na.AbstractNormalRandomSample[CenterT, WidthT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedVectorArray(
    AbstractImplicitVectorArray,
    na.AbstractParameterizedArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorArrayRange(
    AbstractParameterizedVectorArray,
    na.AbstractArrayRange[StartT, StopT],
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorSpace(
    AbstractParameterizedVectorArray,
    na.AbstractSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorLinearSpace(
    AbstractVectorSpace,
    na.AbstractLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorStratifiedRandomSpace(
    AbstractVectorLinearSpace,
    na.AbstractStratifiedRandomSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorLogarithmicSpace(
    AbstractVectorSpace,
    na.AbstractLogarithmicSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractVectorGeometricSpace(
    AbstractVectorSpace,
    na.AbstractGeometricSpace,
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays.tests.test_core
import named_arrays._vectors.cartesian.tests.test_vectors_cartesian_2d
from . import test_matrices_cartesian

__all__ = [
    'AbstractTestAbstractCartesian2dMatrixArray',
    'TestCartesian2dMatrixArray',
]

_num_x = named_arrays.tests.test_core.num_x
_num_y = named_arrays.tests.test_core.num_y
_num_distribution = named_arrays.tests.test_core.num_distribution


class AbstractTestAbstractCartesian2dMatrixArray(
    test_matrices_cartesian.AbstractTestAbstractCartesianMatrixArray,
    named_arrays._vectors.cartesian.tests.test_vectors_cartesian_2d.AbstractTestAbstractCartesian2dVectorArray,
):
    pass


def _cartesian_2d_matrices():
    arrays_xx = [
        4,
        na.ScalarUniformRandomSample(-4, 4, shape_random=dict(y=_num_y)),
    ]
    arrays_xy = [
        1,
    ]
    arrays_yx = [
        5.,
        na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
        na.UniformUncertainScalarArray(
            nominal=na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y)),
            width=1,
            num_distribution=_num_distribution
        )
    ]
    arrays_yy = [0]
    units = [1, u.mm]

    matrices = [
        na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=array_xx, y=array_xy),
            y=na.Cartesian2dVectorArray(x=array_yx, y=array_yy),
        ) * unit
        for array_xx in arrays_xx
        for array_xy in arrays_xy
        for array_yx in arrays_yx
        for array_yy in arrays_yy
        for unit in units
    ]

    matrices.append(
        na.Cartesian2dMatrixArray(
            x=na.Cartesian2dMatrixArray(
                x=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
                y=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
            ),
            y=na.Cartesian2dMatrixArray(
                x=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
                y=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
            ),
        )
    )

    return matrices


def _cartesian_2d_matrices_2():
    arrays_xx = [
        6,
        na.ScalarUniformRandomSample(-6, 6, shape_random=dict(y=_num_y)),
    ]
    arrays_xy = [
        1,
    ]
    arrays_yx = [
        0
    ]
    arrays_yy = [
        7.,
        na.ScalarUniformRandomSample(-7, 7, shape_random=dict(x=_num_x, y=_num_y)),
        na.UniformUncertainScalarArray(
            nominal=na.ScalarUniformRandomSample(-7, 7, shape_random=dict(x=_num_x, y=_num_y)),
            width=1,
            num_distribution=_num_distribution
        )
    ]
    units = [1, u.mm]

    scalars = [
        array_yy * unit
        for array_yy in arrays_yy
        for unit in units
    ]
    vectors = [
        na.Cartesian2dVectorArray(x=array_xx, y=array_yy) * unit
        for array_xx in arrays_xx
        for array_yy in arrays_yy
        for unit in units
    ]
    matrices = [
        na.Cartesian2dMatrixArray(
            x=na.Cartesian2dVectorArray(x=array_xx, y=array_xy),
            y=na.Cartesian2dVectorArray(x=array_yx, y=array_yy),
        ) * unit
        for array_xx in arrays_xx
        for array_xy in arrays_xy
        for array_yx in arrays_yx
        for array_yy in arrays_yy
        for unit in units
    ]

    matrices.append(
        na.Cartesian2dMatrixArray(
            x=na.Cartesian2dMatrixArray(
                x=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
                y=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
            ),
            y=na.Cartesian2dMatrixArray(
                x=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
                y=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(x=1, y=2),
                    y=na.Cartesian2dVectorArray(x=1, y=2)),
            ),
        )
    )


    return scalars + matrices


@pytest.mark.parametrize(
    argnames='array',
    argvalues=_cartesian_2d_matrices()
)
class TestCartesian2dMatrixArray(
    AbstractTestAbstractCartesian2dMatrixArray,
    test_matrices_cartesian.AbstractTestAbstractCartesianMatrixArray,
):

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(y=0),
            dict(y=slice(0, 1)),
            dict(y=na.ScalarArrayRange(0, 2, axis='y')),
            dict(
                y=na.Cartesian2dMatrixArray(
                    x=na.Cartesian2dVectorArray(
                        x=na.ScalarArrayRange(0, 2, axis='y'),
                        y=na.ScalarArrayRange(0, 2, axis='y'),
                    ),
                    y=na.Cartesian2dVectorArray(
                        x=na.ScalarArrayRange(0, 2, axis='y'),
                        y=na.ScalarArrayRange(0, 2, axis='y'),
                    )
                ),
            ),
            dict(
                y=na.UncertainScalarArray(
                    nominal=na.ScalarArray(np.array([0, 1]), axes=('y',)),
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0, ], [1, ]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                ),
                _distribution=na.UncertainScalarArray(
                    nominal=None,
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0], [0]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                )
            ),
            na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
            na.UniformUncertainScalarArray(
                nominal=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                width=0.1,
                num_distribution=_num_distribution,
            ) > 0.5,
            na.Cartesian2dMatrixArray(
                x=na.Cartesian2dVectorArray(
                    x=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.3,
                    y=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.4,
                ),
                y=na.Cartesian2dVectorArray(
                    x=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
                    y=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.6,
                ),
            ),
        ]
    )
    def test__getitem__(
            self,
            array: na.AbstractCartesian2dVectorArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize('array_2', _cartesian_2d_matrices_2())
    class TestUfuncBinary(
        test_matrices_cartesian.AbstractTestAbstractCartesianMatrixArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize('array_2', _cartesian_2d_matrices_2())
    class TestMatmul(
        test_matrices_cartesian.AbstractTestAbstractCartesianMatrixArray.TestMatmul
    ):
        pass


@pytest.mark.parametrize("type_array", [na.Cartesian2dMatrixArray])
class TestCartesian2dMatrixArrayCreation(
    test_matrices_cartesian.AbstractTestAbstractExplicitCartesianMatrixArrayCreation,
):

    @pytest.mark.parametrize("like", [None] + _cartesian_2d_matrices())
    class TestFromScalarArray(
        test_matrices_cartesian.AbstractTestAbstractExplicitCartesianMatrixArrayCreation.TestFromScalarArray,
    ):
        pass



from typing import Callable, Sequence
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays._scalars.uncertainties.uncertainties_array_functions

__all__ = [
    "DEFAULT_FUNCTIONS",
    "PERCENTILE_LIKE_FUNCTIONS",
    "ARG_REDUCE_FUNCTIONS",
    "STACK_LIKE_FUNCTIONS",
    "HANDLED_FUNCTIONS",
    "array_function_default",
    "array_function_percentile_like",
    "array_function_arg_reduce",
    "array_function_stack_like",
    "broadcast_to",
    "tranpose",
    "moveaxis",
    "reshape",
    "array_equal",
]

DEFAULT_FUNCTIONS = named_arrays._scalars.uncertainties.uncertainties_array_functions.DEFAULT_FUNCTIONS
PERCENTILE_LIKE_FUNCTIONS = named_arrays._scalars.uncertainties.uncertainties_array_functions.PERCENTILE_LIKE_FUNCTIONS
ARG_REDUCE_FUNCTIONS = named_arrays._scalars.uncertainties.uncertainties_array_functions.ARG_REDUCE_FUNCTIONS
STACK_LIKE_FUNCTIONS = named_arrays._scalars.uncertainties.uncertainties_array_functions.STACK_LIKE_FUNCTIONS

HANDLED_FUNCTIONS = dict()


def array_function_default(
        func: Callable,
        a: na.AbstractFunctionArray,
        axis: None | str | Sequence[str] = None,
        dtype: None | type | np.dtype = np._NoValue,
        out: None | na.AbstractFunctionArray = None,
        keepdims: bool = True,
        initial: None | bool | int | float | complex | u.Quantity = np._NoValue,
        where: na.AbstractFunctionArray = np._NoValue,
) -> na.FunctionArray:

    a = a.explicit
    inputs = a.inputs
    outputs = a.outputs

    if isinstance(where, na.AbstractArray):
        if isinstance(where, na.AbstractFunctionArray):
            if np.any(where.inputs != inputs):
                raise na.InputValueError("`where.inputs` must match `a.inputs`")
            inputs_where = outputs_where = where.outputs
        elif isinstance(where, (na.AbstractScalar, na.AbstractVectorArray)):
            if where.shape:
                raise ValueError(
                    f"if `where` is an instance of {na.AbstractArray}, but not {na.AbstractFunctionArray}, "
                    f"it must have an empty shape, got {where.shape}"
                )
            inputs_where = outputs_where = where
        else:
            return NotImplemented
    else:
        inputs_where = outputs_where = where

    shape = na.shape_broadcasted(a, where)
    shape_inputs = na.shape_broadcasted(inputs, inputs_where)
    shape_outputs = na.shape_broadcasted(outputs, outputs_where)

    axis_normalized = tuple(shape) if axis is None else (axis,) if isinstance(axis, str) else axis

    if axis is not None:
        if not set(axis_normalized).issubset(shape):
            raise ValueError(
                f"the `axis` argument must be `None` or a subset of the broadcasted shape of `a` and `where`, "
                f"got {axis} for `axis`, but `{shape} for `shape`"
            )

    shape_base = {ax: shape[ax] for ax in axis_normalized}

    kwargs = dict(
        keepdims=keepdims,
    )

    if dtype is not np._NoValue:
        kwargs["dtype"] = dtype
    if initial is not np._NoValue:
        kwargs["initial"] = initial
    if where is not np._NoValue:
        kwargs["where"] = outputs_where

    if isinstance(out, na.AbstractFunctionArray):
        inputs_out = out.inputs
        outputs_out = out.outputs
    else:
        inputs_out = outputs_out = out

    if keepdims:
        if inputs_out is not None:
            np.copyto(src=inputs, dst=inputs_out)
            inputs_result = inputs_out
        else:
            inputs_result = inputs
    else:
        inputs_result = np.mean(
            a=na.broadcast_to(inputs, na.broadcast_shapes(shape_inputs, shape_base)),
            axis=axis_normalized,
            out=inputs_out,
            keepdims=keepdims,
            where=inputs_where,
        )

    outputs_result = func(
        a=na.broadcast_to(outputs, na.broadcast_shapes(shape_outputs, shape_base)),
        axis=axis_normalized,
        out=outputs_out,
        **kwargs,
    )

    if out is None:
        result = a.type_explicit(
            inputs=inputs_result,
            outputs=outputs_result,
        )
    else:
        result = out

    return result


def array_function_percentile_like(
        func: Callable,
        a: na.AbstractFunctionArray,
        q: float | u.Quantity | na.AbstractArray,
        axis: None | str | Sequence[str] = None,
        out: None | na.FunctionArray = None,
        overwrite_input: bool = False,
        method: str = "linear",
        keepdims: bool = False,
) -> na.FunctionArray:

    a = a.explicit
    inputs = a.inputs
    outputs = a.outputs

    shape = a.shape
    shape_inputs = a.inputs.shape
    shape_outputs = a.outputs.shape

    axis_normalized = na.axis_normalized(a, axis)

    if axis is not None:
        if not set(axis_normalized).issubset(shape):
            raise ValueError(
                f"the `axis` argument, {axis}, must be `None` or a subset of the shape of `a`, {shape}"
            )

    shape_base = {ax: shape[ax] for ax in axis_normalized}

    kwargs = dict(
        overwrite_input=overwrite_input,
        method=method,
        keepdims=keepdims,
    )

    if isinstance(out, na.AbstractFunctionArray):
        inputs_out = out.inputs
        outputs_out = out.outputs
    else:
        inputs_out = outputs_out = out

    if keepdims:
        if inputs_out is not None:
            np.copyto(src=inputs, dst=inputs_out)
            inputs_result = inputs_out
        else:
            inputs_result = inputs
    else:
        inputs_result = np.mean(
            a=na.broadcast_to(inputs, na.broadcast_shapes(shape_inputs, shape_base)),
            axis=axis_normalized,
            out=inputs_out,
            keepdims=keepdims,
        )

    outputs_result = func(
        a=na.broadcast_to(outputs, na.broadcast_shapes(shape_outputs, shape_base)),
        q=q,
        axis=axis_normalized,
        out=outputs_out,
        **kwargs,
    )

    if out is None:
        result = a.type_explicit(
            inputs=inputs_result,
            outputs=outputs_result,
        )
    else:
        result = out

    return result


def array_function_arg_reduce(
        func: Callable,
        a: na.AbstractFunctionArray,
        axis: None | str | Sequence[str] = None,
) -> dict[str, na.AbstractArray]:

    return func(a=a.outputs, axis=axis)


def array_function_stack_like(
        func: Callable,
        arrays: Sequence[na.AbstractFunctionArray],
        axis: str,
        out: None | na.FunctionArray = None,
        *,
        dtype: str | np.dtype | type = None,
        casting: str = "same_kind",
):

    if any(not isinstance(array, na.AbstractFunctionArray) for array in arrays):
        return NotImplemented

    if func is np.concatenate:

        if any(axis not in array.shape for array in arrays):
            raise ValueError(f"axis '{axis}' must be present in all the input arrays, got {[a.axes for a in arrays]}")

        arrays_broadcasted = list()
        for array in arrays:
            shape = array.shape
            shape_base = {axis: shape[axis]}
            array = na.FunctionArray(
                inputs=na.broadcast_to(array.inputs, na.broadcast_shapes(array.inputs.shape, shape_base)),
                outputs=na.broadcast_to(array.outputs, na.broadcast_shapes(array.outputs.shape, shape_base)),
            )
            arrays_broadcasted.append(array)
        arrays = arrays_broadcasted

    arrays_inputs = tuple(array.inputs for array in arrays)
    arrays_outputs = tuple(array.outputs for array in arrays)

    if out is None:
        inputs_out = outputs_out = out
    else:
        inputs_out = out.inputs
        outputs_out = out.outputs

    inputs_result = func(
        arrays=arrays_inputs,
        axis=axis,
        out=inputs_out,
    )

    outputs_result = func(
        arrays=arrays_outputs,
        axis=axis,
        out=outputs_out,
        dtype=dtype,
        casting=casting,
    )

    if out is None:
        result = na.FunctionArray(inputs=inputs_result, outputs=outputs_result)
    else:
        out.inputs = inputs_result
        out.outputs = outputs_result
        result = out

    return result


def _implements(numpy_function: Callable):
    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func
    return decorator


@_implements(np.copyto)
def copyto(
        dst: na.FunctionArray,
        src: na.AbstractFunctionArray,
        casting: str = "same_kind",
        where: bool | na.AbstractFunctionArray = True,
) -> None:
    if not isinstance(dst, na.FunctionArray):
        return NotImplemented

    if not isinstance(src, na.AbstractFunctionArray):
        return NotImplemented

    if isinstance(where, na.AbstractArray):
        if isinstance(where, na.AbstractFunctionArray):
            if np.any(where.inputs != src.inputs):
                raise ValueError(f"`where.inputs` must be equivalent to `src.inputs`")
            where_inputs = where.inputs
            where_outputs = where.outputs
        else:
            return NotImplemented
    else:
        where_inputs = where_outputs = where

    try:
        np.copyto(dst=dst.inputs, src=src.inputs, casting=casting, where=where_inputs)
    except TypeError:
        dst.inputs = src.inputs

    try:
        np.copyto(dst=dst.outputs, src=src.outputs, casting=casting, where=where_outputs)
    except TypeError:
        dst.outputs = src.outputs


@_implements(np.broadcast_to)
def broadcast_to(
        array: na.AbstractFunctionArray,
        shape: dict[str, int]
) -> na.FunctionArray:
    return array.type_explicit(
        inputs=na.broadcast_to(array.inputs, shape=shape),
        outputs=na.broadcast_to(array.outputs, shape=shape),
    )


@_implements(np.transpose)
def tranpose(
        a: na.AbstractFunctionArray,
        axes: None | Sequence[str] = None
) -> na.FunctionArray:

    a = a.explicit
    shape = a.shape

    axes_normalized = tuple(reversed(shape) if axes is None else axes)
    #
    # if not set(axes_normalized).issubset(shape):
    #     raise ValueError(f"`axes` {axes} not a subset of `a.axes`, {a.axes}")

    shape_inputs = a.inputs.shape
    shape_outputs = a.outputs.shape

    return a.type_explicit(
        inputs=np.transpose(
            a=a.inputs,
            axes=axes_normalized,
        ),
        outputs=np.transpose(
            a=a.outputs,
            axes=axes_normalized,
        ),
    )


@_implements(np.moveaxis)
def moveaxis(
        a: na.AbstractFunctionArray,
        source: str | Sequence[str],
        destination: str | Sequence[str],
):
    a = a.explicit
    shape = a.shape

    if isinstance(source, str):
        source = source,

    if isinstance(destination, str):
        destination = destination,

    if not set(source).issubset(shape):
        raise ValueError(f"source axes {source} not in array axes {a.axes}")

    shape_inputs = a.inputs.shape
    shape_outputs = a.outputs.shape

    source_destination_inputs = tuple((src, dest) for src, dest in zip(source, destination) if src in shape_inputs)
    source_destination_outputs = tuple((src, dest) for src, dest in zip(source, destination) if src in shape_outputs)

    source_inputs, destination_inputs = tuple(tuple(i) for i in zip(*source_destination_inputs))
    source_outputs, destination_outputs = tuple(tuple(i) for i in zip(*source_destination_outputs))

    return a.type_explicit(
        inputs=np.moveaxis(
            a=a.inputs,
            source=source_inputs,
            destination=destination_inputs,
        ),
        outputs=np.moveaxis(
            a=a.outputs,
            source=source_outputs,
            destination=destination_outputs,
        ),
    )


@_implements(np.reshape)
def reshape(
        a: na.AbstractFunctionArray,
        newshape: dict[str, int],
) -> na.FunctionArray:

    a = np.broadcast_to(a, a.shape)

    return a.type_explicit(
        inputs=np.reshape(a.inputs, newshape=newshape),
        outputs=np.reshape(a.outputs, newshape=newshape)
    )


@_implements(np.array_equal)
def array_equal(
        a1: na.AbstractFunctionArray,
        a2: na.AbstractFunctionArray,
        equal_nan: bool = False,
):
    inputs_equal = np.array_equal(
        a1=a1.inputs,
        a2=a2.inputs,
        equal_nan=equal_nan,
    )
    outputs_equal = np.array_equal(
        a1=a1.outputs,
        a2=a2.outputs,
        equal_nan=equal_nan
    )

    return inputs_equal and outputs_equal


@_implements(np.array_equiv)
def array_equiv(
        a1: na.AbstractFunctionArray,
        a2: na.AbstractFunctionArray,
):
    inputs_equiv = np.array_equiv(
        a1=a1.inputs,
        a2=a2.inputs,
    )
    outputs_equiv = np.array_equiv(
        a1=a1.outputs,
        a2=a2.outputs,
    )

    return inputs_equiv and outputs_equiv


@_implements(np.allclose)
def allclose(
        a: na.AbstractFunctionArray,
        b: na.AbstractFunctionArray,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
):
    close_inputs = np.allclose(
        a=a.inputs,
        b=b.inputs,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan
    )
    close_outputs = np.allclose(
        a=a.outputs,
        b=b.outputs,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
    )
    return close_inputs and close_outputs


@_implements(np.nonzero)
def nonzero(a: na.AbstractFunctionArray) -> dict[str, na.AbstractArray]:
    return np.nonzero(a.outputs)



import pytest
import abc
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    "AbstractTestAbstractTransformation",
]

identities = [
    na.transformations.IdentityTransformation(),
]

translations = [
    na.transformations.Translation(na.Cartesian3dVectorArray(1, 2) * u.mm),
    na.transformations.Translation(na.Cartesian3dVectorLinearSpace(
        start=-10 * u.mm,
        stop=10 * u.mm,
        axis="t",
        num=4,
    )),
]

translations_cartesian_3d = [
    na.transformations.Cartesian3dTranslation(x=1 * u.mm, y=2 * u.mm),
    na.transformations.Cartesian3dTranslation(
        x=na.linspace(-1, 1, axis="t", num=4) * u.mm,
    )
]

transformations_linear = [
    na.transformations.LinearTransformation(na.Cartesian3dIdentityMatrixArray()),
    na.transformations.LinearTransformation(na.Cartesian3dXRotationMatrixArray(43 * u.deg)),
]

angles = [
    53 * u.deg,
    na.linspace(0, 180, axis="angle", num=5) * u.deg
]

rotations_x = [na.transformations.Cartesian3dRotationX(angle) for angle in angles]
rotations_y = [na.transformations.Cartesian3dRotationY(angle) for angle in angles]
rotations_z = [na.transformations.Cartesian3dRotationZ(angle) for angle in angles]


transformations_affine = [
    na.transformations.AffineTransformation(transformation_linear, translation)
    for transformation_linear in transformations_linear
    for translation in translations
]

transformations_basic = identities + translations + transformations_linear + transformations_affine

transformations_list = [
    na.transformations.TransformationList([]),
]

transformations_list += [
    na.transformations.TransformationList([t1, t2], intrinsic=intrinsic)
    for t1 in transformations_basic[::2]
    for t2 in transformations_basic[1::2]
    for intrinsic in (False, True)
]

transformations = transformations_basic + transformations_list


class AbstractTestAbstractTransformation(
    abc.ABC,
):

    @pytest.mark.parametrize(
        argnames="x",
        argvalues=[
            na.Cartesian3dVectorArray(10, 20, 30) * u.mm,
        ]
    )
    class TestVectorOperations:

        def test__call__(
                self,
                a: na.transformations.AbstractTransformation,
                x: na.AbstractVectorArray,
        ):
            result = a(x)
            assert isinstance(result, na.AbstractVectorArray)

        def test_inverse(
                self,
                a: na.transformations.AbstractTransformation,
                x: na.AbstractVectorArray,
        ):
            assert np.allclose(a.inverse(a(x)), x)
            assert np.allclose(a(a.inverse(x)), x)

        @pytest.mark.parametrize("b", transformations)
        def test__matmul__(
                self,
                a: na.transformations.AbstractTransformation,
                b: na.transformations.AbstractTransformation,
                x: na.AbstractVectorArray,
        ):
            assert np.allclose((a @ a)(x), a(a(x)))


@pytest.mark.parametrize("a", identities)
class TestIdentityTransformation(
    AbstractTestAbstractTransformation,
):
    pass


class AbstractTestAbstractTranslation(
    AbstractTestAbstractTransformation,
):
    def test_vector(self, a: na.transformations.AbstractTranslation):
        assert isinstance(a.vector, na.AbstractVectorArray)


@pytest.mark.parametrize("a", translations)
class TestTranslation(
    AbstractTestAbstractTranslation,
):
    pass


@pytest.mark.parametrize("a", translations_cartesian_3d)
class TestCartesian3dTranslation(
    AbstractTestAbstractTranslation
):
    pass


class AbstractTestAbstractLinearTransformation(
    AbstractTestAbstractTransformation,
):
    def test_matrix(self, a: na.transformations.AbstractLinearTransformation):
        assert isinstance(a.matrix, na.AbstractMatrixArray)


@pytest.mark.parametrize("a", transformations_linear)
class TestLinearTransformation(
    AbstractTestAbstractLinearTransformation
):
    pass


class AbstractTestAbstractCartesian3dRotation(
    AbstractTestAbstractLinearTransformation,
):
    pass


@pytest.mark.parametrize("a", rotations_x)
class TestCartesian3dRotationX(
    AbstractTestAbstractCartesian3dRotation,
):
    pass


@pytest.mark.parametrize("a", rotations_y)
class TestCartesian3dRotationY(
    AbstractTestAbstractCartesian3dRotation,
):
    pass


@pytest.mark.parametrize("a", rotations_z)
class TestCartesian3dRotationZ(
    AbstractTestAbstractCartesian3dRotation,
):
    pass


class AbstractTestAbstractAffineTransformation(
    AbstractTestAbstractTransformation,
):
    def test_transformation_linear(self, a: na.transformations.AbstractAffineTransformation):
        assert isinstance(a.transformation_linear, na.transformations.AbstractLinearTransformation)

    def test_translation(self, a: na.transformations.AbstractAffineTransformation):
        assert isinstance(a.translation, na.transformations.Translation)


@pytest.mark.parametrize("a", transformations_affine)
class TestAffineTransformation(
    AbstractTestAbstractAffineTransformation
):
    pass


class AbstractTestAbstractTransformationList(
    AbstractTestAbstractTransformation,
):

    def test_transformations(self, a: na.transformations.AbstractTransformationList):
        for t in a.transformations:
            assert isinstance(t, na.transformations.AbstractTransformation)

    def test_intrinsic(self, a: na.transformations.AbstractTransformationList):
        assert isinstance(a.intrinsic, bool)

    def test__iter__(self, a: na.transformations.AbstractTransformationList):
        for transformation in a:
            assert isinstance(transformation, na.transformations.AbstractTransformation)

    def test_composed(self, a: na.transformations.AbstractTransformationList):
        result = a.composed
        assert isinstance(result, na.transformations.AbstractTransformation)
        assert not isinstance(result, na.transformations.AbstractTransformationList)

    class TestVectorOperations(
        AbstractTestAbstractTransformation.TestVectorOperations
    ):
        def test__call__(
                self,
                a: na.transformations.AbstractTransformationList,
                x: na.AbstractVectorArray,
        ):
            result = a(x)
            result_expected = x
            for transformation in a:
                result_expected = transformation(result_expected)
            assert np.allclose(result, result_expected)


@pytest.mark.parametrize("a", transformations_list)
class TestTransformationList(
    AbstractTestAbstractTransformationList
):
    pass



